#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class provides common statistical metrics for determining how well a source separation algorithm in nussl was
able to create a binary mask compared to a known binary mask. The metrics used here are 
`Precision, Recall <https://en.wikipedia.org/wiki/Precision_and_recall>`_,
`F-Score <https://en.wikipedia.org/wiki/F1_score>`_ (sometimes called F-measure or F1-score), and Accuracy
(though this is not reflected in the name of the class, it is simply   ``# correct / total``).


Example:
    
.. code-block:: python
    :linenos:

    # Run Repet
    repet = nussl.Repet(mixture, mask_type=nussl.BinaryMask)  # it's important to specify BinaryMask!
    repet_masks = repet()
    
    # Get Ideal Binary Masks
    ideal_mask = nussl.IdealMask(mixture, [drums, flute], mask_type=nussl.BinaryMask)  # BinaryMask here, too!
    ideal_masks = ideal_mask()
    
    # Compare Repet to Ideal Binary Mask
    prf_repet = nussl.PrecisionRecallFScore(ideal_masks, repet_masks)
    prf_repet_scores = prf_repet.evaluate()

Scores for each source are stored in a nested dictionary aptly named ``scores``. This is a dictionary of dictionaries
where the key is the source label, and the value is another dictionary with scores for each of the metrics for that
source. So, for instance, the format of the ``prf_repet_scores`` dictionary from above is shown below:

.. code-block:: python

    {'Source 0' : {'Accuracy': 0.83,
                   'Precision': 0.78,
                   'Recall': 0.81,
                   'F1-Score': 0.77 },
     'Source 1' : {'Accuracy': 0.22,
                   'Precision': 0.12,
                   'Recall': 0.15,
                   'F1-Score': 0.19 }
    }


Notes:
    * ``PrecisionRecallFScore`` can only be run using :ref:`binary_mask` objects. The constructor expects a list of 
    :ref:`binary_mask` objects for both the ground truth sources and the estimated sources.
    * ``PrecisionRecallFScore`` does not calculate the correct permutation of the estimated and ground truth sources;
    they are expected to be in the correct order when they are passed into ``PrecisionRecallFScore``.

See Also:
    * :ref:`evaluation_base` for more information about derived properties that this class has.
    
    * :ref:`ideal_mask` for information about how to get an array of ground truth binary masks.

"""

import sklearn

import evaluation_base
from ..separation.masks import binary_mask


class PrecisionRecallFScore(evaluation_base.EvaluationBase):
    """
    Args:
        true_sources_mask_list (list): List of :ref:`binary_mask` objects representing the ground truth sources.
        estimated_sources_mask_list (list): List of :ref:`binary_mask` objects representing the estimates from a source
         separation object
        source_labels (list) (Optional): List of ``str`` with labels for each source. If no labels are provided, sources
         will be labeled ``Source 0, Source 1, ...`` etc.
         
    Attributes:
        scores (dict): Dictionary storing the precision, recall, F1-Score, and accuracy. 
         See :ref:`nussl.PrecisionRecallFScore.evaluate` below.
        
    """

    ACCURACY_KEY = 'Accuracy'
    PRECISION_KEY = 'Precision'
    RECALL_KEY = 'Recall'
    FSCORE_KEY = 'F1-Score'

    def __init__(self, true_sources_mask_list, estimated_sources_mask_list, source_labels=None):
        super(PrecisionRecallFScore, self).__init__(true_sources_list=true_sources_mask_list,
                                                    estimated_sources_list=estimated_sources_mask_list,
                                                    source_labels=source_labels)

    @staticmethod
    def _verify_input_list(mask_list):
        if isinstance(mask_list, binary_mask.BinaryMask):
            mask_list = [mask_list]
        elif isinstance(mask_list, list):
            if not all(isinstance(m, binary_mask.BinaryMask) for m in mask_list):
                raise ValueError('All items in mask_list must be of type nussl.BinaryMask!')

        if not all(mask_list[0].shape == m.shape for m in mask_list):
            raise ValueError('All masks must be the same shape!')

        return mask_list

    @staticmethod
    def _preprocess(mask1, mask2):
        """
        Prepares masks for sklearn metric functions. Both ``mask1`` and ``mask2`` must be ``BinaryMask`` objects. 
        Args:
            mask1 (:obj:`BinaryMask`): BinaryMask
            mask2 (:obj:`BinaryMask`): BinaryMask

        Returns:
            [mask1, mask2] ready to be input to into an sklearn metric.

        """
        assert isinstance(mask1, binary_mask.BinaryMask)
        assert isinstance(mask2, binary_mask.BinaryMask)
        return mask1.mask.ravel(), mask2.mask.ravel()

    def _precision(self, true_mask, estimated_mask):
        """
        Wrapper for sklearn.metrics.precision_score()
        Args:
            true_mask: 
            estimated_mask: 

        Returns:

        """

        return sklearn.metrics.precision_score(*self._preprocess(true_mask, estimated_mask))

    def _recall(self, true_mask, estimated_mask):
        """
        Wrapper for sklearn.metrics.recall_score()
        Args:
            true_mask: 
            estimated_mask: 

        Returns:

        """
        return sklearn.metrics.recall_score(*self._preprocess(true_mask, estimated_mask))

    def _f_score(self, true_mask, estimated_mask):
        """
        Warpper for sklearn.metrics.f1_score()
        Args:
            true_mask: 
            estimated_mask: 

        Returns:

        """
        return sklearn.metrics.f1_score(*self._preprocess(true_mask, estimated_mask))

    def _accuracy(self, true_mask, estimated_mask):
        """
        Wrapper for sklearn.metrics.accuracy_score()
        Args:
            true_mask: 
            estimated_mask: 

        Returns:

        """
        return sklearn.metrics.accuracy_score(*self._preprocess(true_mask, estimated_mask))

    def evaluate(self):
        """
        Determines the precision, recall, f-score, and accuracy of each :ref:`binary_mask` object in 
        ``true_sources_mask_list`` and ``estimated_sources_mask_list``. Returns a dictionary of results that is
        formatted like so:
        
        .. code-block:: python

            {'Source 0' : {'Accuracy': 0.83,
                           'Precision': 0.78,
                           'Recall': 0.81,
                           'F1-Score': 0.77 },
             'Source 1' : {'Accuracy': 0.22,
                           'Precision': 0.12,
                           'Recall': 0.15,
                           'F1-Score': 0.19 }
            }
        
        This dictionary is stored as e keys to this dictionary 
        
        Returns:
            self.scores (dict): A dictionary of scores that contains accuracy, precision, recall, and F1-score
            of between the list of :ref:`binary_mask` objects in both ``true_sources_mask_list`` 
            and ``estimated_sources_mask_list``.

        """
        for i, true_mask in enumerate(self.true_sources_list):
            est_mask = self.estimated_sources_list[i]

            label = self.source_labels[i]
            results = {self.ACCURACY_KEY: self._accuracy(true_mask, est_mask),
                       self.PRECISION_KEY: self._precision(true_mask, est_mask),
                       self.RECALL_KEY: self._recall(true_mask, est_mask),
                       self.FSCORE_KEY: self._f_score(true_mask, est_mask)}

            self.scores[label] = results

        return self.scores



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
While *nussl* does not come with any data sets, it does have the capability to interface with
many common source separation data sets used within the MIR and speech separation communities.
These data set "hooks" are implemented as generator functions, and allow the user to loop through
each file in the data set and get separation :class:`AudioSignal` objects for the mixture and
individual sources.
"""
import os
import hashlib
import warnings

import numpy as np

import efz_utils
from audio_signal import AudioSignal

__all__ = ['iKala', 'mir1k', 'timit', 'medleyDB', 'musdb18', 'dsd100']


def _hash_directory(directory, ext=None):
    """
    Calculates the hash of every child file in the given directory using python's built-in SHA256
    function (using `os.walk()`, which also searches subdirectories recursively). If :param:`ext`
    is specified, this will only look at files with extension provided.

    This function is used to verify the integrity of data sets for use with nussl. Pretty much
    just makes sure that when we loop through/look at a directory, we understand the structure
    because the organization of the data set directories for different data sets are all unique
    and thus need to be hard coded by each generator function (below). If we get a hash mismatch
    we can throw an error easily.

    Args:
        directory (str): Directory within which file hashes get calculated. Searches recursively.
        ext (str): If provided, this function will only calculate the hash on files with the given
            extension.

    Returns:
        (str): String containing only hexadecimal digits of the has of the
            contents of the given directory.

    """

    hash_list = []
    for path, sub_dirs, files in os.walk(directory):
        if ext is None:
            hash_list.extend([efz_utils._hash_file(os.path.join(path, f)) for f in files
                              if os.path.isfile(os.path.join(path, f))])
        else:
            hash_list.extend([efz_utils._hash_file(os.path.join(path, f)) for f in files
                              if os.path.isfile(os.path.join(path, f))
                              if os.path.splitext(f)[1] == ext])

    hasher = hashlib.sha256()
    for hash_val in sorted(hash_list):  # Sort this list so we're platform agnostic
        hasher.update(hash_val.encode('utf-8'))

    return hasher.hexdigest()


def _check_hash(audio_dir, check_hash, expected_hash, ext):
    """
    Checks to see if the hashed contents of :param:`audio_dir` match :param:`expected_hash`. If
    :param:`check_hash` is ``False``, this function does nothing. If :param:`check_hash` is
    ``'warn'`` (a string) and there is a hash mismatch, then nussl will print a warning
    message to the console. If :param:`check_hash` is ``True`` and there is a mismatch, this
    function will throw a :class:`DataSetException`.

    Args:
        audio_dir (str): Top level directory to check the hash.
        check_hash (bool, str): In the case that there is a mismatch between the expected and
            calculated hash, if this parameter is ``True`` (a bool) an exception is raised and
            if this parameter is ``'warn'`` (a string) a warning is printed to the console. If
            this parameter is ``False``, the hash will not be calculated for this directory, i.e.,
            this function does nothing.
        expected_hash (str): The known, pre-computed hash that this function will test against.
        ext (str): When calculating the hash, only look at files with the provided extension.

    """
    # Check to see if the contents are what we expect
    # by hashing all of the files and checking against a precomputed expected_hash
    if check_hash and _hash_directory(audio_dir, ext) != expected_hash:
        msg = 'Hash of {} does not match known directory hash!'.format(audio_dir)
        if check_hash == 'warn':
            warnings.warn(msg)
        else:
            raise DataSetException(msg)


def _data_set_setup(directory, top_dir_name, audio_dir_name, expected_hash, check_hash, ext):
    """
    This function does some preliminary checking to make sure the directory we get from the user
    is oriented how we expect, checks the hash of the directory (if :param:`check_hash` is ``True``
    or ``'warn'``), and then spits out a list of the full path names of every audio file.

    Args:
        directory (str): The directory containing of the directory of the data set.
        top_dir_name (str): Name of the top-level directory of the data set. This should be the one
            that is downloaded from the data set host.
        audio_dir_name (str): Name of the directory that contains the audio files.
        expected_hash (str): Pre-calculated hash that this function will test against.
        check_hash (bool, str): In the case that there is a mismatch between the expected and
            calculated hash, if this parameter is ``True`` (a bool) an exception is raised and
            if this parameter is ``'warn'`` (a string) a warning is printed to the console. If
            this parameter is ``False``, the hash will not be calculated for this directory, i.e.,
            this function does nothing.
        ext (str): When calculating the hash, only look at files with the provided extension.

    Returns:
        (list): List of full paths for every audio file in the data set directory.
    """

    # Verify the top-level directory is correct
    if not os.path.isdir(directory):
        raise DataSetException('Expected directory, got \'{}\'!'.format(directory))

    if top_dir_name != os.path.split(directory)[1]:
        raise DataSetException('Expected {}, got \'{}\''.format(top_dir_name,
                                                                os.path.split(directory)[1]))

    # This should be tha directory with all of the audio files
    audio_dir = os.path.join(directory, audio_dir_name)
    if not os.path.isdir(audio_dir):
        raise DataSetException('Expected {} to be a directory but it is not!'.format(audio_dir))

    _check_hash(audio_dir, check_hash, expected_hash, ext)

    # return a list of the full paths of every audio file
    return [os.path.join(audio_dir, f) for f in os.listdir(audio_dir)
            if os.path.isfile(os.path.join(audio_dir, f))
            if os.path.splitext(f)[1] == ext]


def _subset_and_shuffle(file_list, subset, shuffle, seed):
    """
    Given a list of file paths, this function will shuffle and make subsets of the given list.

    Args:
        file_list (list): List of full paths of audio files.
        subset (float, list, str, None): This parameter determines how to make a subset of the
        audio files in the data set. There are four ways to use it, depending on what type
        this parameter takes:
        1) If :param:`subset` is a ``float``, then :param:`subset` will return the first
           ``X.Y%`` of audio files, where ``X.Y%`` is some arbitrary percentage. In this case,
           :param:`subset` is expected to be in the range [0.0, 1.0].
        2) If :param:`subset` is a ``list``, it is expected to be a list of indices (as
           ``int``s). This function will then produce the audio files in the list that correspond
           to those indices.
        3) If :param:`subset` is a ``str``, it will only include audio files with that string
          somewhere in the directory name.
        4) If :param:`subset` is ``None``, then the whole data set is traversed unobstructed.

        shuffle (bool): Whether the data set should be shuffled.
        seed (int, 1-d array_like): Seed for ``numpy``'s random number generator used for
            shuffling.

    Returns:
        (list): Resultant list of file paths that is either shuffled and/or subset according to the
        input parameters.

    """

    if shuffle:
        if seed is not None:
            np.random.seed(seed)

        np.random.shuffle(file_list)

    if isinstance(subset, float):
        if subset < 0.0 or subset > 1.0:
            raise DataSetException('subset must be a number between [0.0, 1.0]!')

        result = file_list[:int(subset*len(file_list))]

    elif isinstance(subset, list):
        result = [file_list[i] for i in subset]

    elif isinstance(subset, (str, bytes)):
        result = [f for f in file_list if subset in os.path.dirname(f)]

    elif subset is None:
        result = file_list

    else:
        raise DataSetException('subset must be a list of indices or float between [0.0, 1.0]!')

    return result


def iKala(directory, check_hash=True, subset=None, shuffle=False, seed=None):
    """
    Generator function for the iKala data set. This allows you to loop through the entire data set
    with only a few :class:`AudioSignal` objects stored in memory at a time. There are options for
    only looping through a subset of the data set and shuffling the data set (with a seed). See
    details about those options below.

    `nussl` calculates the hash of the iKala directory and compares it against a precomputed hash
    for iKala that ships with `nussl`. This hash is used to verify that `nussl` understands the
    directory structure when reading the files. Calculating the hash can be turned off if the
    user needs a speed up, but this might cause oblique errors if the iKala directory is not set up
    in the same way as a fresh download of iKala.

    Examples:
        Using this generator function to loop through the iKala data set. In this example, we use
        the generator directly in the ``for`` loop.

        .. code-block:: python
            :linenos:

            iKala_path = '/path/to/iKala'  # the iKala directory in disc
            for mix, vox, acc in nussl.datasets.iKala(iKala_path):
                mix.to_mono(overwrite=True)  # sum to mono to make a 'mixture'

                # Get some basic metadata on the files.
                # (They'll all have the same file name, but different labels)
                print('Mixture       - Filename: {}, Label: {}'.format(mix.file_name, mix.label))
                print('Vocals        - Filename: {}, Label: {}'.format(vox.file_name, vox.label))
                print('Accompaniment - Filename: {}, Label: {}'.format(acc.file_name, acc.label))

                # Run an algorithm on the iKala files and save to disc
                r = nussl.Repet(mix)
                r.run()
                bg_est, fg_est = r.make_audio_signals()
                bg_est.write_audio_to_file('{}_bg.wav'.format(os.path.splitext(mix.file_name)[0]))
                fg_est.write_audio_to_file('{}_fg.wav'.format(os.path.splitext(mix.file_name)[0]))

        It's also possible to use ``tqdm`` to print the progress to the console. This is useful
        because running through an entire data set can take a while. Here's a more advanced example
        using some other options as well:

        .. code-block:: python
            :linenos:

            import nussl
            import tdqm

            iKala_path = 'path/to/iKala' # the iKala directory on disc
            idxs = range(29, 150)[::2]  # Only get every other song between [29, 150)
            iKala_gen = nussl.datasets.iKala(iKala_path, subset=idxs, check_hash=False)

            # Tell tqdm the number of files we're running on so it can estimate a completion time
            for mixture, vocals, accompaniment in tqdm(iKala_gen, total=len(idxs)):
                mix.to_mono(overwrite=True)  # sum to mono to make a 'mixture'

                # Run an algorithm on the iKala files and save to disc
                r = nussl.Repet(mix)
                r.run()
                bg_est, fg_est = r.make_audio_signals()
                bg_est.write_audio_to_file('{}_bg.wav'.format(os.path.splitext(mix.file_name)[0]))
                fg_est.write_audio_to_file('{}_fg.wav'.format(os.path.splitext(mix.file_name)[0]))

    Args:
        directory (str): Top-level directory for the iKala data set.
        check_hash (bool, str): In the case that there is a mismatch between the expected and
        calculated hash, if this parameter is ``True`` (a bool) an exception is raised and
        if this parameter is ``'warn'`` (a string) a warning is printed to the console. If
        this parameter is ``False``, the hash will not be calculated for this directory, i.e.,
        this function does nothing.
        subset (float, list, str, None): This parameter determines how to make a subset of the
        audio files in the data set. There are four ways to use it, depending on what type
        this parameter takes:
        1) If :param:`subset` is a ``float``, then :param:`subset` will return the first
           ``X.Y%`` of audio files, where ``X.Y%`` is some arbitrary percentage. In this case,
           :param:`subset` is expected to be in the range [0.0, 1.0].
        2) If :param:`subset` is a ``list``, it is expected to be a list of indices (as
           ``int``s). This function will then produce the audio files in the list that correspond
           to those indices.
        3) If :param:`subset` is a ``str``, it will only include audio files with that string
         somewhere in the directory name.
        4) If :param:`subset` is ``None``, then the whole data set is traversed unobstructed.
        shuffle (bool): Whether the data set should be shuffled.
        seed (int, 1-d array_like): Seed for ``numpy``'s random number generator used for
        shuffling.

    Yields:
        (``tuple(AudioSignal, AudioSignal, AudioSignal)``):
            A tuple of three :class:`AudioSignal` objects, with audio loaded for each source. In
            the tuple, they are returned in the following order:
            ``(mixture, vocals, accompaniment)``. In iKala, the audio files are such that the
            vocals are hard panned to one channel and the accompaniment is hard panned to the other.
            So, the 'mixture' yielded here by this function reflects this, and needs to 'mixed'
            down to mono. In other words, ``mixture`` is a stereo :class:`AudioSignal` object,
            where each channel is on source, and similarly ``vocals`` and ``accompaniment`` are
            mono :class:`AudioSignal` objects made from a single channel in `mixture`.

    """

    top_dir_name = 'iKala'
    audio_dir_name = 'Wavfile'
    iKala_hash = 'd82191c73c3ce0ab0ed3ca21a3912769394e8c9a16d742c6cc446e3f04a9cd9e'
    audio_extension = '.wav'
    all_wav_files = _data_set_setup(directory, top_dir_name, audio_dir_name,
                                    iKala_hash, check_hash, audio_extension)

    all_wav_files = _subset_and_shuffle(all_wav_files, subset, shuffle, seed)

    for f in all_wav_files:
        mixture = AudioSignal(f)
        mixture.label = 'mixture'

        vocals = mixture.make_audio_signal_from_channel(1)
        vocals.label = 'vocals'

        accompaniment = mixture.make_audio_signal_from_channel(0)
        accompaniment.label = 'accompaniment'

        yield mixture, vocals, accompaniment


def mir1k(directory, check_hash=True, subset=None, shuffle=False, seed=None, undivided=False):
    """
    Generator function for the MIR-1K data set. This allows you to loop through the entire data set
    with only a few :class:`AudioSignal` objects stored in memory at a time. There are options for
    only looping through a subset of the data set and shuffling the data set (with a seed). See
    details about those options below.

    `nussl` calculates the hash of the MIR-1K directory and compares it against a precomputed hash
    for MIR-1K that ships with `nussl`. This hash is used to verify that `nussl` understands the
    directory structure when reading the files. Calculating the hash can be turned off if the
    user needs a speed up, but this might cause oblique errors if the MIR-1K directory is not set up
    in the same way as a fresh download of MIR-1K.

    MIR-1K also ships with two 'sets' of audio files: the divided and undivided sets. They contain
    the same content, the only difference is that the undivided set is one file per song, each song
    taking up the whole file, and the divided set has the same song divided into segments of ~3-12
    seconds. The :param:`undivided` parameter controls which of these two sets `nussl` will loop
    through.

    Examples:
        Using this generator function to loop through the MIR-1K data set. In this example, we use
        the generator directly in the ``for`` loop.

        .. code-block:: python
            :linenos:

            mir1k_path = '/path/to/MIR-1K'  # the MIR-1K directory in disc
            for mix, vox, acc in nussl.datasets.mir1k(mir1k_path):
                mix.to_mono(overwrite=True)  # sum to mono to make a 'mixture'

                # Get some basic metadata on the files.
                # (They'll all have the same file name, but different labels)
                print('Mixture       - Filename: {}, Label: {}'.format(mix.file_name, mix.label))
                print('Vocals        - Filename: {}, Label: {}'.format(vox.file_name, vox.label))
                print('Accompaniment - Filename: {}, Label: {}'.format(acc.file_name, acc.label))

                # Run an algorithm on the MIR-1K files and save to disc
                r = nussl.Repet(mix)
                r.run()
                bg_est, fg_est = r.make_audio_signals()
                bg_est.write_audio_to_file('{}_bg.wav'.format(os.path.splitext(mix.file_name)[0]))
                fg_est.write_audio_to_file('{}_fg.wav'.format(os.path.splitext(mix.file_name)[0]))

        It's also possible to use ``tqdm`` to print the progress to the console. This is useful
        because running through an entire data set can take a while. Here's a more advanced example
        using some other options as well:

        .. code-block:: python
            :linenos:

            import nussl
            import tdqm

            mir1k_path = 'path/to/MIR-1K' # the MIR-1K directory on disc
            idxs = range(29, 150)[::2]  # Only get every other song between [29, 150)
            mir1k_gen = nussl.datasets.mir1k(mir1k_path, subset=idxs,
                                             check_hash=False, undivided=True)

            # Tell tqdm the number of files we're running on so it can estimate a completion time
            for mixture, vocals, accompaniment in tqdm(mir1k_gen, total=len(idxs)):
                mix.to_mono(overwrite=True)  # sum to mono to make a 'mixture'

                # Run an algorithm on the MIR-1K files and save to disc
                r = nussl.Repet(mix)
                r.run()
                bg_est, fg_est = r.make_audio_signals()
                bg_est.write_audio_to_file('{}_bg.wav'.format(os.path.splitext(mix.file_name)[0]))
                fg_est.write_audio_to_file('{}_fg.wav'.format(os.path.splitext(mix.file_name)[0]))

    Args:
        directory (str): Top-level directory for the MIR-1K data set.
        check_hash (bool, str): In the case that there is a mismatch between the expected and
        calculated hash, if this parameter is ``True`` (a bool) an exception is raised and
        if this parameter is ``'warn'`` (a string) a warning is printed to the console. If
        this parameter is ``False``, the hash will not be calculated for this directory, i.e.,
        this function does nothing.
        subset (float, list, str, None): This parameter determines how to make a subset of the
        audio files in the data set. There are four ways to use it, depending on what type
        this parameter takes:
        1) If :param:`subset` is a ``float``, then :param:`subset` will return the first
        ``X.Y%`` of audio files, where ``X.Y%`` is some arbitrary percentage. In this case,
        :param:`subset` is expected to be in the range [0.0, 1.0].
        2) If :param:`subset` is a ``list``, it is expected to be a list of indices (as
        ``int``s). This function will then produce the audio files in the list that correspond
        to those indices.
        3) If :param:`subset` is a ``str``, it will only include audio files with that string
        somewhere in the directory name.
        4) If :param:`subset` is ``None``, then the whole data set is traversed unobstructed.
        shuffle (bool): Whether the data set should be shuffled.
        seed (int, 1-d array_like): Seed for ``numpy``'s random number generator used for
        shuffling.
        undivided (bool): Whether to use the divided (in the ``Wavefile`` directory) or undivided
        (in the ``UndividedWavefile`` directory).

    Yields:
        (``tuple(AudioSignal, AudioSignal, AudioSignal)``):
            A tuple of three :class:`AudioSignal` objects, with audio loaded for each source. In
            the tuple, they are returned in the following order:
            ``(mixture, vocals, accompaniment)``. In MIR-1K, the audio files are such that the
            vocals are hard panned to one channel and the accompaniment is hard panned to the other.
            So, the 'mixture' yielded here by this function reflects this, and needs to 'mixed'
            down to mono. In other words, ``mixture`` is a stereo :class:`AudioSignal` object,
            where each channel is on source, and similarly ``vocals`` and ``accompaniment`` are
            mono :class:`AudioSignal` objects made from a single channel in `mixture`.

    """

    top_dir_name = 'MIR-1K'

    wavfile_hash = '33c085c1a7028199cd20317868849b413e0971022ebc4aefcf1bbc5516646c29'
    undivided_hash = '3f39af9be17515e042a7005b4c47110c6738623a7fada6233ba104535b7dde1b'

    if undivided:
        audio_dir_name = 'UndividedWavfile'
        mir1k_hash = undivided_hash
    else:
        audio_dir_name = 'Wavfile'
        mir1k_hash = wavfile_hash

    audio_extension = '.wav'
    all_wav_files = _data_set_setup(directory, top_dir_name, audio_dir_name,
                                    mir1k_hash, check_hash, audio_extension)

    all_wav_files = _subset_and_shuffle(all_wav_files, subset, shuffle, seed)

    for f in all_wav_files:
        mixture = AudioSignal(f)
        mixture.label = 'mixture'

        vocals = mixture.make_audio_signal_from_channel(1)
        vocals.label = 'vocals'

        accompaniment = mixture.make_audio_signal_from_channel(0)
        accompaniment.label = 'accompaniment'

        yield mixture, vocals, accompaniment


def timit(directory, check_hash=True, subset=None, shuffle=False, seed=None):
    """
    Not implemented yet.

    Args:
        directory:
        check_hash:
        subset:
        shuffle:
        seed:

    Yields:

    """
    dir_name = 'TIMIT'
    hash = ''
    audio_extension = '.wav'
    all_wav_files = _data_set_setup(directory, dir_name, dir_name,
                                    hash, check_hash, audio_extension)

    all_wav_files = _subset_and_shuffle(all_wav_files, subset, shuffle, seed)


def medleyDB(directory, raw=False, check_hash=True, subset=None, shuffle=False, seed=None):
    """
    Not implemented yet.

    Args:
        directory:
        check_hash:
        subset:
        shuffle:
        seed:

    Returns:

    """
    top_dir_name = 'medleyDB'
    audio_dir_name = 'Audio'
    hash = ''
    audio_extension = '.wav'


def dsd100(directory, check_hash=True, subset=None, shuffle=False, seed=None):
    """
    Not implemented yet.

    Args:
        directory:
        check_hash:
        subset:
        shuffle:
        seed:

    Returns:

    """


def musdb18(directory, check_hash=True, subset=None, folder=None, shuffle=False, seed=None):
    """
    Generator function for the MUSDB18 data set. This allows you to loop through the entire data set
    with only a few :class:`AudioSignal` objects stored in memory at a time. There are options for
    only looping through a subset of the data set and shuffling the data set (with a seed). See
    details about those options below.

    `nussl` calculates the hash of the MUSDB18 directory and compares it against a precomputed hash
    for MUSDB18 that ships with `nussl`. This hash is used to verify that `nussl` understands the
    directory structure when reading the files. Calculating the hash can be turned off if the
    user needs a speed up, but this might cause oblique errors if the MUSDB directory is not set up
    in the same way as a fresh download of MUSDB18.

    The audio in MUSDB18 is stored in the 'stempeg' format from Native Instruments. `nussl` uses
    the `stempeg` library to read these files from disc, and returns each of the sources as
    individual :obj:`AudioSignal` objects.

    Examples:
        Using this generator function to loop through the MUSDB18 data set. In this example, we use
        the generator directly in the ``for`` loop.

        .. code-block:: python
            :linenos:

            musdb_path = '/path/to/MUSDB18'  # the MUSDB18 directory in disc
            for mix, drums, bass, other, vox in nussl.datasets.musdb(musdb_path):

                # Get some basic metadata on the files.
                # (They'll all have the same file name, but different labels)
                print('Mixture  - Filename: {}, Label: {}'.format(mix.file_name, mix.label))
                print('Vocals   - Filename: {}, Label: {}'.format(vox.file_name, vox.label))
                print('Drums    - Filename: {}, Label: {}'.format(drums.file_name, drums.label))

                # Run an algorithm on the MUSDB18 files and save to disc
                r = nussl.Repet(mix)
                r.run()
                bg_est, fg_est = r.make_audio_signals()
                bg_est.write_audio_to_file('{}_bg.wav'.format(os.path.splitext(mix.file_name)[0]))
                fg_est.write_audio_to_file('{}_fg.wav'.format(os.path.splitext(mix.file_name)[0]))

        It's also possible to use ``tqdm`` to print the progress to the console. This is useful
        because running through an entire data set can take a while. Here's a more advanced example
        using some other options as well:

        .. code-block:: python
            :linenos:

            import nussl
            import tdqm

            musdb_path = 'path/to/MUSDB18' # the MUSDB18 directory on disc

            # Only run on the 'test' folder (this has 50 songs)
            musdb_gen = nussl.datasets.musdb(musdb_path, subset='test', check_hash=False)

            # Tell tqdm the number of files we're running on so it can estimate a completion time
            for mix, drums, bass, other, vox in tqdm(musdb_gen, total=50):

                # Run an algorithm on the MUSDB18 files and save to disc
                r = nussl.Repet(mix)
                r.run()
                bg_est, fg_est = r.make_audio_signals()
                bg_est.write_audio_to_file('{}_bg.wav'.format(os.path.splitext(mix.file_name)[0]))
                fg_est.write_audio_to_file('{}_fg.wav'.format(os.path.splitext(mix.file_name)[0]))

    Args:
        directory (str): Top-level directory for the MUSDB18 data set.
        check_hash (bool, str): In the case that there is a mismatch between the expected and
            calculated hash, if this parameter is ``True`` (a bool) an exception is raised and
            if this parameter is ``'warn'`` (a string) a warning is printed to the console. If
            this parameter is ``False``, the hash will not be calculated for this directory, i.e.,
            this function does nothing.
        subset (float, list, str, None): This parameter determines how to make a subset of the
        audio files in the data set. There are four ways to use it, depending on what type
        this parameter takes:
        * If :param:`subset` is a ``float``, then :param:`subset` will return the first
        ``X.Y%`` of audio files, where ``X.Y%`` is some arbitrary percentage. In this case,
        :param:`subset` is expected to be in the range [0.0, 1.0].
        ( If :param:`subset` is a ``list``, it is expected to be a list of indices (as
        ``int``s). This function will then produce the audio files in the list that correspond
        to those indices.
        ( If :param:`subset` is a ``str``, it will only include audio files with that string
        somewhere in the directory name.
        * If :param:`subset` is ``None``, then the whole data set is traversed unobstructed.
        shuffle (bool): Whether the data set should be shuffled.
        seed (int, 1-d array_like): Seed for ``numpy``'s random number generator used for
        shuffling.

    Yields:
        (tuple):
            A tuple of five :class:`AudioSignal` objects, with audio loaded for each source. In
            the tuple, they are returned in the following order:
            ``(mixture, drums, bass, other, vox)``.
    """

    try:
        import stempeg
    except Exception:
        raise ImportError('Cannot read MUSDB18 without stempeg package installed!')

    dir_name = 'musdb18'
    audio_dir_names = ['test', 'train'] if folder not in ('test', 'train') else [folder]
    hash = 'cf4cfcef4eadc212c34df6e8fb1184a3f63b7fedfab23e79d17e735fff0bfaf9'
    audio_extension = '.mp4'
    stem_labels = ['mixture', 'drums', 'bass', 'other', 'vocals']

    # Check the hash has of the full directory
    _check_hash(directory, check_hash, hash, audio_extension)

    files = []
    for audio_dir_name in audio_dir_names:
        files.extend(_data_set_setup(directory, dir_name, audio_dir_name,
                                     '', False, audio_extension))

    files = _subset_and_shuffle(files, subset, shuffle, seed)

    for f in files:
        stem, sr = stempeg.read_stems(f)
        yield tuple(_make_audio_signal_from_stem(f, stem, i, sr, l)
                    for i, l in enumerate(stem_labels))


def _make_audio_signal_from_stem(filename, stem, i, sr, label):
    """
    Reads in a :param:`stem` from the stempeg library (`stempeg.read_stems()`) and creates a
    correctly formatted :class:`AudioSignal` object (with all the metadata set up).
    Args:
        filename (str): Name of the file on disc.
        stem (:obj:`np.ndarray`): Numpy array from the `stempeg.read_stems()` function.
        i (int): Index of the :param:`stem: array to get audio data from.
        sr (int): Sample rate.
        label (str): Label for the :class:`AudioSignal` object.

    Returns:
        (:obj:`AudioSignal`) Correctly formatted :class:`AudioSignal` object with the
            right metadata.

    """
    signal = AudioSignal(audio_data_array=stem[i,...], sample_rate=sr)
    signal.path_to_input_file = filename
    signal.label = label
    return signal


class DataSetException(Exception):
    """
    Exception class for errors when working with data sets in nussl.
    """
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

from ..core import constants
import masks
import mask_separation_base
from .. import AudioSignal


class RPCA(mask_separation_base.MaskSeparationBase):
    """Implements foreground/background separation using RPCA

    Huang, Po-Sen, et al. "Singing-voice separation from monaural recordings using robust
    principal component analysis." Acoustics, Speech and Signal Processing (ICASSP), 2012
    IEEE International Conference on. IEEE, 2012.

    Parameters:
        input_audio_signal: (AudioSignal object) The AudioSignal object that has the
                            audio data that RPCA will be run on.
        high_pass_cutoff: (Optional) (float) value (in Hz) for the high pass cutoff filter.
        num_iterations: (Optional) (int) how many iterations to run RPCA for, defaults to 100.
        verbose: (Optional) (bool) prints out error rates and iteration count while running (defaults to False)
        do_mono: (Optional) (bool) Flattens AudioSignal to mono before running the algorithm (does not effect the
                        input AudioSignal object)
        use_librosa_stft: (Optional) (bool) Calls librosa's stft function instead of nussl's

    """

    def __init__(self, input_audio_signal, high_pass_cutoff=None, num_iterations=None, epsilon=None,
                 do_mono=False, verbose=False, use_librosa_stft=constants.USE_LIBROSA_STFT,
                 mask_type=mask_separation_base.MaskSeparationBase.SOFT_MASK, mask_threshold=0.5):
        super(RPCA, self).__init__(input_audio_signal=input_audio_signal, mask_type=mask_type)
        self.high_pass_cutoff = 100.0 if high_pass_cutoff is None else float(high_pass_cutoff)
        self.use_librosa_stft = use_librosa_stft

        self.epsilon = 1e-7 if epsilon is None else epsilon
        self.num_iterations = 100 if num_iterations is None else num_iterations
        self.gain = 1
        self.verbose = verbose

        self.error = None
        self.stft = None
        self.magnitude_spectrogram = None
        self.background = None
        self.foreground = None
        self.result_masks = None

        if do_mono:
            self.audio_signal.to_mono(overwrite=True)

    def run(self):
        """

        Returns:
            background (AudioSignal): An AudioSignal object with repeating background in background.audio_data
            (to get the corresponding non-repeating foreground run self.make_audio_signals())

        Example:
             ::

        """
        # High pass filter cutoff freq. (in # of freq. bins), +1 to match MATLAB implementation
        self.high_pass_cutoff = int(
            np.ceil(self.high_pass_cutoff * (self.stft_params.n_fft_bins - 1) /
                    self.audio_signal.sample_rate)) + 1

        self._compute_spectrum()

        # separate the mixture background by masking
        background_stft = []
        background_mask = []
        for i in range(self.audio_signal.num_channels):
            background = self.compute_rpca_mask(self.magnitude_spectrogram[:, :, i])
            background[0:self.high_pass_cutoff, :] = 1  # high-pass filter the foreground
            background_mask.append(background)

            # apply mask
            stft_with_mask = background * self.stft[:, :, i]
            background_stft.append(stft_with_mask)

        background_stft = np.array(background_stft).transpose((1, 2, 0))
        self.background = AudioSignal(stft=background_stft,
                                      sample_rate=self.audio_signal.sample_rate)
        self.background.istft(self.stft_params.window_length, self.stft_params.hop_length,
                              self.stft_params.window_type,
                              overwrite=True, use_librosa=self.use_librosa_stft,
                              truncate_to_length=self.audio_signal.signal_length)

        background_mask = np.array(background_mask).transpose((1, 2, 0)).astype('float')
        background_mask = masks.SoftMask(background_mask)
        if self.mask_type == self.BINARY_MASK:
            background_mask = background_mask.mask_to_binary(self.mask_threshold)

        self.result_masks = [background_mask, background_mask.inverse_mask()]

        return self.result_masks

    def _compute_spectrum(self):
        self.stft = self.audio_signal.stft(overwrite=True, remove_reflection=True,
                                           use_librosa=self.use_librosa_stft)
        self.magnitude_spectrogram = np.abs(self.stft)

    def compute_rpca_mask(self, magnitude_spectrogram):
        low_rank, sparse_matrix = self.decompose(magnitude_spectrogram)
        bg_mask = self.gain * np.abs(sparse_matrix) <= np.abs(low_rank)
        return bg_mask

    def decompose(self, magnitude_spectrogram):
        # compute rule of thumb values of lagrange multiplier and svd-threshold
        _lambda = 1 / np.sqrt(np.max(magnitude_spectrogram.shape))

        # initialize low rank and sparse matrices to all zeros
        low_rank = np.zeros(magnitude_spectrogram.shape)
        sparse_matrix = np.zeros(magnitude_spectrogram.shape)

        # get singular values for magnitude_spectrogram
        two_norm = np.linalg.svd(magnitude_spectrogram, full_matrices=False, compute_uv=False)[0]
        inf_norm = np.linalg.norm(magnitude_spectrogram.flatten(), np.inf) / _lambda
        dual_norm = np.max([two_norm, inf_norm])
        residuals = magnitude_spectrogram / dual_norm

        # tunable parameters
        mu = 1.25 / two_norm
        mu_bar = mu * 1e7
        rho = 1.5

        error = np.inf
        converged = False
        num_iteration = 0

        while not converged and num_iteration < self.num_iterations:
            if self.verbose:
                print('Iteration: {}, Error: {}'.format(num_iteration, error))

            num_iteration += 1
            low_rank = self.svd_threshold(magnitude_spectrogram - sparse_matrix + residuals / mu,
                                          1 / mu)
            sparse_matrix = self.shrink(magnitude_spectrogram - low_rank + residuals / mu,
                                        _lambda / mu)
            residuals += mu * (magnitude_spectrogram - low_rank - sparse_matrix)
            mu = np.min([mu * rho, mu_bar])
            error = np.linalg.norm(magnitude_spectrogram - low_rank - sparse_matrix,
                                   ord='fro') / np.linalg.norm(magnitude_spectrogram, ord='fro')
            if error < self.epsilon:
                converged = True
        self.error = error
        return low_rank, sparse_matrix

    def shrink(self, matrix, tau):
        return np.sign(matrix) * np.maximum(np.abs(matrix) - tau, 0)

    def svd_threshold(self, matrix, tau):
        u, sigma, v = np.linalg.svd(matrix, full_matrices=False)
        shrunk = self.shrink(sigma, tau)
        thresholded_singular_values = np.dot(u, np.dot(np.diag(shrunk), v))
        return thresholded_singular_values

    def reduced_rank_svd(self, matrix, k):
        u, sigma, v = np.linalg.svd(matrix, full_matrices=False)
        matrix_reduced = np.dot(u[:, 0:k], np.dot(sigma[0:k], v[0:k, :]))
        return matrix_reduced

    def make_audio_signals(self):
        """ Returns the background and foreground audio signals. You must have run FT2D.run() prior
        to calling this function. This function will return None if run() has not been called.

        Returns:
            Audio Signals (List): 2 element list.

                * bkgd: Audio signal with the calculated background track
                * fkgd: Audio signal with the calculated foreground track

        EXAMPLE:
             ::
        """
        if self.background is None:
            raise ValueError('Cannot make audio signals prior to running algorithm!')

        foreground_array = self.audio_signal.audio_data - self.background.audio_data
        self.foreground = self.audio_signal.make_copy_with_audio_data(foreground_array)
        return [self.background, self.foreground]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Deep Clustering Separation Class
"""
import copy
import warnings

try:
    import torch
    from torch.autograd import Variable
    torch_okay = True
except ImportError:
    warnings.warn('Cannot import pytorch!')
    torch_okay = False

from sklearn.cluster import KMeans
import librosa
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

from ..transformers import TransformerDeepClustering
from sklearn.decomposition import PCA
import mask_separation_base
import masks


class DeepClustering(mask_separation_base.MaskSeparationBase):
    """Implements deep clustering for source separation, using PyTorch.

    Deep clustering is a deep learning approach to source separation. It takes as input a
    mel-spectrogram representation of an audio mixture. Each time-frequency bin is mapped into
    an K-dimensional embedding. The model works out so that time-frequency bins that are dominated
    by different sources map to embeddings that are distant, and bins that are dominated by the
    same source map to embeddings that are near. The sources are then recovered using K-Means
    clustering on the embedding space.

    References:

    Hershey, J. R., Chen, Z., Le Roux, J., & Watanabe, S. (2016, March).
    Deep clustering: Discriminative embeddings for segmentation and separation.
    In Acoustics, Speech and Signal Processing (ICASSP),
    2016 IEEE International Conference on (pp. 31-35). IEEE.

    Luo, Y., Chen, Z., Hershey, J. R., Roux, J. L., & Mesgarani, N. (2016).
    Deep Clustering and Conventional Networks for Music Separation: Stronger Together.
    arXiv preprint arXiv:1611.06265.

    Example:
        music = AudioSignal("path/to/input.wav", offset=45, duration=20)

        music.stft_params.window_length = 2048
        music.stft_params.hop_length = 512

        separation = DeepClustering(music, num_sources = 2)
        masks = separation.run()
        sources = separation.make_audio_signals()
        plt.figure(figsize=(20, 8))
        separation.plot()
        plt.tight_layout()
        plt.show()
    """
    def __init__(self, input_audio_signal,
                 mask_type=mask_separation_base.MaskSeparationBase.SOFT_MASK,
                 model_path='/media/ext/models/deep_clustering_vocal_44k_long.model',
                 num_sources=2,
                 num_layers=4,
                 hidden_size=500,
                 max_distance=1,
                 embedding_size=20,
                 num_mels=150,
                 do_mono=False,
                 resample_rate=44100,
                 use_librosa_stft=False,
                 cutoff=-40):

        if not torch_okay:
            raise ImportError('Cannot import pytorch! Install pytorch to continue.')

        super(DeepClustering, self).__init__(input_audio_signal=input_audio_signal,
                                             mask_type=mask_type)

        self.resample_rate = resample_rate
        if self.audio_signal.sample_rate != self.resample_rate:
            self.audio_signal.resample(self.resample_rate)

        self.use_librosa_stft = use_librosa_stft
        self.num_mels = num_mels
        self.num_sources = num_sources
        self.num_fft = self.audio_signal.stft_params.n_fft_bins
        self.mel_filter_bank = librosa.filters.mel(self.resample_rate,
                                                   self.num_fft, self.num_mels).T
        self.inverse_mel_filter_bank = np.linalg.pinv(self.mel_filter_bank)
        self.max_distance = max_distance
        self.use_cuda = torch.cuda.is_available()
        if not self.use_cuda:
            warnings.warn('CUDA not available, running on CPU may be slow')

        self.stft = None
        self.mel_spectrogram = None
        self.silence_mask = None
        self.sources = None
        self.assignments = None
        self.centroids = None
        self.masks = []
        self.cutoff = cutoff
        self.model = TransformerDeepClustering(num_layers=num_layers,
                                               hidden_size=hidden_size,
                                               embedding_size=embedding_size)
        if self.use_cuda:
            self.model = self.model.cuda()

        self.load_model(model_path)
        self.clusterer = KMeans(n_clusters=self.num_sources)
        self.embeddings = None

        self.do_mono = do_mono

        if self.do_mono:
            self.audio_signal.to_mono(overwrite=True)

    def load_model(self, model_path):
        """
        Loads the model at specified path ``model_path``
        Args:
            model_path:

        Returns:

        """
        self.model.load_state_dict(torch.load(model_path,
                                              map_location=lambda storage, loc: storage))
        self.model.eval()

    def _compute_spectrograms(self):
        self.stft = self.audio_signal.stft(overwrite=True, remove_reflection=True,
                                           use_librosa=self.use_librosa_stft)
        magnitude = np.abs(self.stft)
        self.mel_spectrogram = np.empty((self.audio_signal.num_channels,
                                         self.stft.shape[1], self.num_mels))

        for i in range(self.audio_signal.num_channels):
            self.mel_spectrogram[i, :, :] = np.dot(magnitude[:, :, i].T, self.mel_filter_bank)

        self.mel_spectrogram = 10.0 * np.log10(self.mel_spectrogram**2 + 1e-7)
        self.silence_mask = self.mel_spectrogram > self.cutoff
        self.mel_spectrogram -= np.mean(self.mel_spectrogram)
        self.mel_spectrogram /= np.std(self.mel_spectrogram) + 1e-7

    def deep_clustering(self):
        """

        Returns:

        """
        input_data = Variable(torch.FloatTensor(self.mel_spectrogram))

        if self.use_cuda:
            input_data = input_data.cuda()

        if self.embeddings is None:
            embeddings = self.model(input_data)
            self.embeddings = embeddings.view(-1, embeddings.size(-1)).cpu().data.numpy()
        self.clusterer.fit(self.embeddings)

        assignments = self.clusterer.labels_ + 1
        self.assignments = assignments.reshape(self.mel_spectrogram.shape)
        self.centroids = self.clusterer.cluster_centers_

    def _extract_masks(self, ch):
        if self.audio_signal.stft_data is None:
            raise ValueError('Cannot extract masks with no signal_stft data')

        channel_mask_list = []

        for cluster_index in range(1, self.num_sources + 1):
            if self.mask_type == self.BINARY_MASK:
                mask = ((self.silence_mask[ch, :, :] * self.assignments[ch, :, :]) == cluster_index)

            elif self.mask_type == self.SOFT_MASK:
                d = (self.embeddings - self.centroids[cluster_index-1])**2
                distances = np.sqrt(np.sum(d, axis=-1))
                distances += distances.min()
                distances /= (distances.max() + 1e-7)
                distances[distances > self.max_distance] = 1
                distances = 1 - distances.reshape(self.mel_spectrogram.shape)
                mask = (self.silence_mask[ch, :, :] * distances[ch, :, :])
            else:
                raise ValueError('Unknown mask type {}!'.format(self.mask_type))

            mask = np.dot(mask, self.inverse_mel_filter_bank).T
            mask += np.abs(mask.min())
            mask /= (np.max(mask) + 1e-7)
            channel_mask_list.append(mask)

        if self.mask_type == self.SOFT_MASK:
            data = np.stack(channel_mask_list, axis=-1)
            original_shape = data.shape
            data = data.reshape(-1, 2)**3
            data = (data.T / np.sum(data, axis=-1)).T
            data = data.reshape(original_shape)
            channel_mask_list = [data[:, :, i] for i in range(self.num_sources)]

        return channel_mask_list

    def generate_mask(self, ch, assignments):
        """
            Takes binary Mel spectrogram assignments and generates mask
        """
        if self.audio_signal.stft_data is None:
            raise ValueError('Cannot extract masks with no signal_stft data')

        mask = (self.silence_mask[ch, :, :] * assignments)
        mask = np.dot(mask, self.inverse_mel_filter_bank).T
        mask += np.abs(mask.min())
        mask /= (np.max(mask) + 1e-7)
        mask = np.round(mask)

        # mask = np.dstack([mask, mask])

        return masks.BinaryMask(mask)

    def run(self):
        """

        Returns:

        """
        self._compute_spectrograms()
        self.deep_clustering()

        uncollated_masks = []
        for i in range(self.audio_signal.num_channels):
            uncollated_masks += self._extract_masks(i)

        collated_masks = [np.dstack([uncollated_masks[s + ch * self.num_sources]
                                     for ch in range(self.audio_signal.num_channels)])
                          for s in range(self.num_sources)]

        self.masks = []

        for mask in collated_masks:
            if self.mask_type == self.BINARY_MASK:
                mask = np.round(mask)
                mask_object = masks.BinaryMask(mask)
            elif self.mask_type == self.SOFT_MASK:

                mask_object = masks.SoftMask(mask)
            else:
                raise ValueError('Unknown mask type {}!'.format(self.mask_type))
            self.masks.append(mask_object)
        return self.masks

    def apply_mask(self, mask):
        """
            Applies individual mask and returns audio_signal object
        """
        source = copy.deepcopy(self.audio_signal)
        source = source.apply_mask(mask)
        source.stft_params = self.stft_params
        source.istft(overwrite=True, truncate_to_length=self.audio_signal.signal_length)

        return source

    def make_audio_signals(self):
        """ Applies each mask in self.masks and returns a list of audio_signal
         objects for each source.
        Returns:
            self.sources (np.array): An array of audio_signal objects
            containing each separated source
        """
        self.sources = []
        for mask in self.masks:
            self.sources.append(self.apply_mask(mask))

        return self.sources

    def project_embeddings(self, num_dimensions):
        """
        Does a PCA projection of the embedding space
        Args:
            num_dimensions:

        Returns:

        """
        transform = PCA(n_components=num_dimensions)
        output_transform = transform.fit_transform(self.embeddings)
        return output_transform

    def plot(self, **kwargs):
        """ Plots relevant information for deep clustering onto the active figure,
            given by matplotlib.pyplot.figure()
            outside of this function. The three plots are:
                1. PCA of emeddings onto 2 dimensions for visualization
                2. The mixture mel-spectrogram.
                3. The source assignments of each tf-bin in the mixture spectrogram.
        Returns:
            None
        """
        grid = GridSpec(6, 10)
        output_transform = self.project_embeddings(2)
        plt.subplot(grid[:3, 3:])
        plt.imshow(np.mean(self.mel_spectrogram, axis=0).T, origin='lower',
                   aspect='auto', cmap='magma')
        plt.xticks([])
        plt.ylabel('Frequency (mel)')
        plt.title('Mixture')

        plt.subplot(grid[1:-1, :3])

        xmin = output_transform[:, 0].min()
        xmax = output_transform[:, 0].max()
        ymin = output_transform[:, 1].min()
        ymax = output_transform[:, 1].max()

        plt.hexbin(output_transform[:, 0], output_transform[:, 1], bins='log', gridsize=100)
        plt.axis([xmin, xmax, ymin, ymax])
        plt.xlabel('PCA dim 1')
        plt.ylabel('PCA dim 2')
        plt.title('Embedding visualization')

        plt.subplot(grid[3:, 3:])
        plt.imshow(np.max(self.silence_mask * self.assignments, axis=0).T,
                   origin='lower', aspect='auto', cmap='Greys')
        plt.xlabel('Time (frames)')
        plt.ylabel('Frequency (mel)')
        plt.title('Source assignments')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Init for ``nussl``, the Northwestern University Source Separation Library.
"""

try:
    import torch
    torch_imported = True
except Exception:
    torch_imported = False

try:
    import vamp
    vamp_imported = True
except Exception:
    vamp_imported = False


class ImportErrorClass(object):
    def __init__(self, lib, **kwargs):
        msg = 'Cannot import {} because {} is not installed'.format(type(self).__name__, lib)
        raise ImportError(msg)


from .core.constants import *
from .core.audio_signal import AudioSignal
from .core import utils, efz_utils, stft_utils, datasets
from .evaluation import *
from .separation import *
from .transformers import *
import core.constants

__all__ = ['core', 'utils', 'stft_utils', 'transformers', 'separation', 'evaluation']


__version__ = '0.1.6'

version = __version__  # aliasing version
short_version = '.'.join(version.split('.')[:-1])

__title__ = 'nussl'
__description__ = 'A flexible sound source separation library.'
__uri__ = 'https://github.com/interactiveaudiolab/nussl'

__author__ = 'E. Manilow, P. Seetharaman, F. Pishdadian'
__email__ = 'ethanmanilow@u.northwestern.edu'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2017 Interactive Audio Lab'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Base class for both BSS Eval algorithms (:ref:`BSSEvalSources` and :ref:`BSSEvalImages`). Contains
most of the logic for these base classes.
"""

import numpy as np

import evaluation_base


class BSSEvalBase(evaluation_base.EvaluationBase):
    """
    Base class for ``mir_eval`` implementation of the BSS-Eval metrics (SDR, SIR, SAR).
    Contains logic for loading ground truth :class:`AudioSignal`s and estimated
    :class:`AudioSignal`s to compute BSS-Eval metrics. The ``mir_eval`` module contains
    an implementation of BSS-Eval version 3.

    The BSS-Eval metrics attempt to measure perceptual quality by comparing sources
    estimated from a source separation algorithm to the ground truth, known sources.
    These metrics evaluate the distortion (SDR) and artifacts (SAR) present in the
    estimated signals as well as the interference (SIR) from other sources in a given
    estimated source. Results are returned in units of dB, with higher values indicating
    better quality.

    See Also:
        * For more information on ``mir_eval`` (python implementation of BSS-Eval v3) see
        `its Github page<https://github.com/craffel/mir_eval>`.
        * For more information on the BSS-Eval metrics, see the webpage for
        `the original MATLAB implementation<http://bass-db.gforge.inria.fr/bss_eval/>`.
        * Implementations of this base class: :class:`BSSEvalSources` and :class:`BSSEvalImages`.
        * :class:`BSSEvalV4` for the ``museval`` version 4 BSS-Eval implementation.

    References:
        * Emmanuel Vincent, Rémi Gribonval, Cédric Févotte. Performance measurement in blind
        audio source separation. IEEE Transactions on Audio, Speech and Language Processing,
        Institute of Electrical and Electronics Engineers, 2006, 14 (4), pp.1462–1469.
        <inria-00544230>
        * Colin Raffel, Brian McFee, Eric J. Humphrey, Justin Salamon, Oriol Nieto, Dawen Liang,
        and Daniel P. W. Ellis, "mir_eval: A Transparent Implementation of Common MIR Metrics",
        Proceedings of the 15th International Conference on Music Information Retrieval, 2014.

    Args:
        true_sources_list (list): List of :class:`AudioSignal` objects that contain the ground
            truth sources for the mixture.
        estimated_sources_list (list):  List of :class:`AudioSignal` objects that contain estimate
            sources, output from source separation algorithms.
        source_labels (list): List of strings that are labels for each source to be used as keys for
            the scores. Default value is ``None`` and in that case labels are ``Source 0``,
            ``Source 1``, etc.
        algorithm_name (str): Name of the algorithm if using this object to compute many
            BSS-Eval metrics. Can be changed later.
        do_mono (bool): Should flatten the audio to mono before calculating metrics.
        compute_permutation (bool): Should try to find the best permutation for the estimated
            sources.

    """
    SDR = 'SDR'
    SIR = 'SIR'
    SAR = 'SAR'
    ISR = 'ISR'
    PERMUTATION = 'permutation'
    RAW_VALUES = 'raw_values'

    def __init__(self, true_sources_list, estimated_sources_list, source_labels=None,
                 algorithm_name=None, do_mono=False, compute_permutation=True):
        super(BSSEvalBase, self).__init__(true_sources_list=true_sources_list,
                                          estimated_sources_list=estimated_sources_list,
                                          source_labels=source_labels, do_mono=do_mono)

        if algorithm_name is None:
            self._algorithm_name = 'Approach'
        else:
            assert type(algorithm_name) == str
            self._algorithm_name = algorithm_name

        self.compute_permutation = compute_permutation
        self._mir_eval_func = None

    @property
    def algorithm_name(self):
        """
        Name of the algorithm that is being evaluated.
        Returns:
            (str) Name of the algorithm being evaluated.
        """
        return self._algorithm_name

    @algorithm_name.setter
    def algorithm_name(self, value):
        assert type(value) == str
        self._algorithm_name = value
        self.scores[self._algorithm_name] = {label: {} for label in self.source_labels}

    def validate(self):
        """
        Checks to make sure the all of the input :class:`AudioSignal` objects have the
        same length.
        """
        # TODO: This might be obsolete
        if self.estimated_sources_list is None:
            raise BssEvalException('Must set estimated_sources first!')
        estimated_lengths = [x.signal_length for x in self.estimated_sources_list]
        reference_lengths = [x.signal_length for x in self.true_sources_list]

        if len(set(estimated_lengths)) > 1:
            raise BssEvalException('All AudioSignals in estimated_sources must be the same length!')
        if len(set(reference_lengths)) > 1:
            raise BssEvalException('All AudioSignals in ground_truth must be the same length!')
    
    def _preprocess_sources(self):
        """
        Prepare the :ref:`audio_data` in the sources for ``mir_eval``.
        Returns:
            (:obj:`np.ndarray`, :obj:`np.ndarray`) reference_source_array, estimated_source_array

        """
        estimated_source_array = np.vstack([np.copy(x.audio_data)
                                            for x in self.true_sources_list])
        reference_source_array = np.vstack([np.copy(x.audio_data)
                                            for x in self.estimated_sources_list])

        return reference_source_array, estimated_source_array

    def evaluate(self):
        """
        Actually runs the evaluation algorithm. Will be ``museval.metrics.bss_eval_images`` or
        ``museval.metrics.bss_eval_sources`` depending on which subclass is instantiated.
        Returns:
            (dict): Dictionary containing the scores.

        """
        self.validate()
        reference, estimated = self._preprocess_sources()

        if self._mir_eval_func is None:
            raise NotImplementedError('Cannot call base class! Try calling '
                                      'BSSEvalSources or BSSEvalImages')

        bss_output = self._mir_eval_func(reference, estimated,
                                         compute_permutation=self.compute_permutation)

        self._populate_scores_dict(bss_output)

        return self.scores

    def _populate_scores_dict(self, bss_output):
        """
        Formats and populates the :attr:`scores` dict from :func:`evaluate`.
        Args:
            bss_output (tuple): Direct output from the ``museval`` function.

        Returns:
            (dict) Reformatted dictionary from ``museval`` output.
        """


class BssEvalException(Exception):
    """
    Exception class for BSS-Eval
    """
    pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division

import warnings

import numpy as np

import separation_base
from ..core import stft_utils
from ..core import constants

from ft2d import FT2D
from repet import Repet
from repet_sim import RepetSim


class OverlapAdd(separation_base.SeparationBase):
    """
    Implements foreground/background separation using overlap/add with an arbitrary foreground/background
        separation scheme in nussl.

    Notes:
        Currently supports ``Repet``, ``RepetSim``, and ``FT2D``.

    Parameters:
        input_audio_signal (:class:`audio_signal.AudioSignal`): The :class:`audio_signal.AudioSignal` object that the 
        OverlapAdd algorithm will be run on. This makes a copy of ``input_audio_signal``

        separation_method:
        overlap_window_size:
        overlap_hop_size:
        overlap_window_type:
        do_mono:
        use_librosa_stft:

    Example:

    .. code-block:: python
        :linenos:
        
         import nussl
         
         signal = nussl.AudioSignal('path/to/audio.wav')
         
         ola = nussl.OverlapAdd(signal, nussl.Repet)  # initialize with class
         ola = nussl.OverlapAdd(signal, 'repet')  # initialize with string (case insensitive)
         ola.run()
         
    """
    def __init__(self, input_audio_signal, separation_method,
                 overlap_window_size=24, overlap_hop_size=12, overlap_window_type=constants.WINDOW_TRIANGULAR,
                 do_mono=False, use_librosa_stft=constants.USE_LIBROSA_STFT):
        super(OverlapAdd, self).__init__(input_audio_signal=input_audio_signal)
        self.background = None
        self.foreground = None

        self.use_librosa_stft = use_librosa_stft
        self.overlap_window_size = overlap_window_size
        self.overlap_hop_size = overlap_hop_size
        self.overlap_window_type = overlap_window_type

        self.window_samples = int(np.round(self.audio_signal.sample_rate * self.overlap_window_size))
        self.hop_samples = int(np.round(self.audio_signal.sample_rate * self.overlap_hop_size))
        self.overlap_samples = self.window_samples - self.hop_samples

        # Set the separation method
        self._separation_method = None
        self._separation_instance = None

        self.separation_method = separation_method

        if do_mono:
            self.audio_signal.to_mono(overwrite=True)

    # Here is where we store valid separation methods for OverlapAdd.
    # To add a new method for OverlapAdd, follow these four steps:
    #
    # 0) import the class at the top of this file.
    # 1) Add the class object here (not an instance), as below:
    repet = Repet
    repet_sim = RepetSim
    ft2d = FT2D

    # 2) And populate this dictionary with your new method like so:
    _valid_separation_methods = {Repet.__name__: repet,
                                 RepetSim.__name__: repet_sim,
                                 FT2D.__name__: ft2d}

    # 3) Add it to the tests in tests/test_overlap_add.py. If you don't, I will hunt you down!!!

    @staticmethod
    def valid_separation_methods():
        """
        Returns a list of class objects that OverlapAdd can use as its separation method
        Returns:

        """
        return OverlapAdd._valid_separation_methods.values()

    @staticmethod
    def valid_separation_method_names():
        """
        Returns a list of strings of class names that OverlapAdd can use as its separation method
        Returns:

        """
        return [method.__name__ for method in OverlapAdd._valid_separation_methods.values()]

    @staticmethod
    def _valid_methods_lower():
        """Case invariant (lowercase) version of self._valid_separation_methods dictionary.
        """
        return {OverlapAdd._format(name): obj for name, obj in OverlapAdd._valid_separation_methods.items()}

    @property
    def separation_method_name(self):
        """
        Returns the name of the current separation object
        Returns:

        """
        return str(self.separation_instance)

    @property
    def separation_method(self):
        """
        Returns the current separation object
        Returns:

        """
        return self._separation_method

    @separation_method.setter
    def separation_method(self, value):
        """
        Sets self.separation_method to value
        Resets self._separation_instance to None
        Args:
            value:

        Returns:

        """
        error = ValueError("Invalid separation method for OverlapAdd! \n" +
                           "Got {0}, but valid methods are: {1}"
                           .format(value, ', '.join(self._valid_separation_methods.keys())))
        if value is None:
            raise error

        if isinstance(value, str):
            if self._format(value) in self._valid_methods_lower().keys():
                # The user input a string with a valid method name. It should be in our dictionary
                self._separation_method = self._valid_methods_lower()[self._format(value)]
            else:
                # Oops. Can't find it in our dictionary
                raise error

        elif issubclass(value, separation_base.SeparationBase) and \
                        value in self._valid_separation_methods.values():
            # The user gave us a class, so we use that
            self._separation_method = value

        else:
            raise error

        self._setup_instance()

    @staticmethod
    def _format(string):
        """ Formats a class name correctly for self._valid_methods_lower.
            Strips all non-alphanumeric chars and makes lowercase.
        """
        return str(filter(str.isalnum, string)).lower()

    @property
    def separation_instance(self):
        """ This the actual instance of the separation algorithm. If you need to make any modifications to
        the default behavior of the algorithm, this is where you would do it.

        Returns:
            (:obj:`SeparationBase`) instance of the separation method.
        Examples:
            ::


        """
        return self._separation_instance

    def _setup_instance(self):
        # instantiate the separation method here
        if self.separation_method is None:
            raise Exception('Cannot separate before separation_method is set!')

        self._separation_instance = self.separation_method(self.audio_signal, use_librosa_stft=self.use_librosa_stft)

    def __str__(self):
        name = super(OverlapAdd, self).__str__()
        name += ':' + self.separation_method.__name__ if self.separation_method is not None else ''
        return name

    def run(self):
        """

        Returns:
            background (:obj:`AudioSignal`): An AudioSignal object with background in background.audio_data
            (to get the corresponding non-repeating foreground run self.make_audio_signals())

        Example:
             ::

        """
        if self._separation_instance is None:
            self._setup_instance()

        # if our window is larger than the total number of samples in the file,
        # just run the algorithm like normal
        if self.audio_signal.signal_length < self.window_samples + self.hop_samples:
            warnings.warn('input_audio_signal length is less than one window. '
                          'Running {} normally...'.format(self.separation_method_name))

            self._separation_instance.run()
            self.background, _ = self._separation_instance.make_audio_signals()
            return self.background

        background_array = np.zeros_like(self.audio_signal.audio_data)

        # Make the window for multiple channels
        window = stft_utils.make_window(self.overlap_window_type, 2 * self.overlap_samples)
        window = np.vstack([window for _ in range(self.audio_signal.num_channels)])

        # Main overlap-add loop
        for start, end in self._next_window():

            if start == 0:
                # First window is a partial window
                first_window = window[:, -self.hop_samples:]
                unwindowed = self._set_active_region_and_run(start, self.hop_samples)
                background_array[:, :self.hop_samples] = np.multiply(unwindowed.audio_data, first_window)

            elif end >= self.audio_signal.signal_length:
                # Last window is a partial window
                remaining = self.audio_signal.signal_length - start
                last_window = window[:, remaining:] if remaining != window.shape[-1] else window
                last_window[:, self.overlap_samples:] = 1  # only do part of the window

                unwindowed = self._set_active_region_and_run(start, self.audio_signal.signal_length)
                background_array[:, start:] += np.multiply(unwindowed.audio_data, last_window)

            else:
                # middle cases are straight forward
                unwindowed = self._set_active_region_and_run(start, end)
                background_array[:, start:end] += np.multiply(unwindowed.audio_data, window)

        self.audio_signal.set_active_region_to_default()
        self.background = self.audio_signal.make_copy_with_audio_data(background_array, verbose=False)
        return self.background

    def _next_window(self):
        """
        Generator that calculates the start and end sample indices for the next window.
        Yields:

        """
        n_segments = 1 + int((self.audio_signal.signal_length-self.window_samples) / self.hop_samples)

        # We went to return values larger than the signal length so we know when the last
        # n_segments = 1 + int( self.audio_signal.signal_length / self.hop_samples)

        for segment in range(n_segments):
            start = segment * self.hop_samples
            end = start + self.window_samples

            yield start, end

    def _set_active_region_and_run(self, start, end):
        self._separation_instance.audio_signal.set_active_region(start, end)
        self._separation_instance.run()
        bkgnd, _ = self._separation_instance.make_audio_signals()
        return bkgnd


    def make_audio_signals(self):
        """ Returns the background and foreground audio signals. You must have run :func:`run()` prior
        to calling this function. This function will raise ValueError if :func:`run()` has not been called.

        Returns:
            Audio Signals (List): 2 element list.

                * bkgd: Audio signal with the calculated background track
                * fkgd: Audio signal with the calculated foreground track

        EXAMPLE:
             ::
        """
        if self.background is None:
            raise ValueError('Cannot make audio signals prior to running algorithm!')

        foreground_array = self.audio_signal.audio_data - self.background.audio_data
        self.foreground = self.audio_signal.make_copy_with_audio_data(foreground_array)
        return [self.background, self.foreground]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import numpy as np
import nussl

import nimfa


class TransformerNMFUnitTests(unittest.TestCase):

    n_attempts = 2
    max_error_pct = 0.05
    n_bases = 2
    n_iters = 100

    def test_small_mixture_matrix(self):
        """
        A simple example of NMF using euclidean and divergence in nussl with a 4 by 4 matrix
        """
        # Make two simple matrices
        for n in range(4, 20, 4):
            a = np.arange(n ** 2).reshape((n, n))
            b = np.add(np.multiply(2., a), 3.)

            # Mix them together
            mixture = np.dot(b, a)

            # Run with euclidean distance
            dist_type = nussl.transformers.TransformerNMF.EUCLIDEAN
            self.calculate_nmf_error(mixture, self.n_bases, dist_type, self.n_iters, self.n_attempts, n)

            # Run with divergence
            dist_type = nussl.transformers.TransformerNMF.KL_DIVERGENCE
            self.calculate_nmf_error(mixture, self.n_bases, dist_type, self.n_iters, self.n_attempts, n)


    def test_random_matrix(self):
        for n in range(4, 10, 2):
            matrix = np.random.rand(n, n)

            # Run on euclidean
            distance_type = nussl.transformers.TransformerNMF.EUCLIDEAN
            self.calculate_nmf_error(matrix, self.n_bases, distance_type, self.n_iters, self.n_attempts, n)

            # Run on divergence
            distance_type = nussl.transformers.TransformerNMF.KL_DIVERGENCE
            self.calculate_nmf_error(matrix, self.n_bases, distance_type, self.n_iters, self.n_attempts, n)

    def calculate_nmf_error(self, mixture, n_bases, dist_type, iterations, attempts, seed):
        div = nussl.transformers.TransformerNMF.KL_DIVERGENCE
        nimfa_type = 'divergence' if dist_type == div else dist_type

        for i in range(attempts):
            # Set up nussl NMF
            nussl_nmf = nussl.TransformerNMF(mixture, n_bases, max_num_iterations=iterations,
                                             distance_measure=dist_type, seed=seed)
            # Run nussl NMF
            nussl_nmf.transform()

            # Set up nimfa NMF
            nimfa_nmf = nimfa.Nmf(mixture, max_iter=iterations, rank=n_bases, update=nimfa_type,
                                  W=nussl_nmf.template_dictionary,
                                  H=nussl_nmf.activation_matrix)  # init to same matrices as nussl

            # Run nimfa NMF
            nmf_fit = nimfa_nmf()

            # Dot the results
            nimfa_est = np.dot(nmf_fit.basis(), nmf_fit.coef())
            nussl_est = np.dot(nussl_nmf.template_dictionary, nussl_nmf.activation_matrix)

            # calculate errors
            max_nussl_error = np.max(np.abs(nussl_est - mixture) / mixture)
            max_nimfa_error = np.max(np.abs(nimfa_est - mixture) / mixture)
            max_diff = max_nussl_error - max_nimfa_error

            # IF nussl's max error is bigger than nimfa's
            # AND nussl's max error bigger than the specified max error (0.05, or 5%)
            # AND the difference between the max errors is larger than 0.05
            # THEN we throw an exception
            # i.e., make sure nussl's results are close to nimfa's
            if max_nussl_error > max_nimfa_error \
                    and max_nussl_error > self.max_error_pct \
                    and max_diff > self.max_error_pct:
                raise Exception('max nussl error is larger than nimfa and self.max_error_pct')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

"""

import museval

import bss_eval_base


class BSSEvalImages(bss_eval_base.BSSEvalBase):
    """
    Wrapper class for ``mir_eval`` implementation of the BSS-Eval Iamges metrics (SDR, SIR, SAR).
    Contains logic for loading ground truth AudioSignals and estimated
    AudioSignals to compute BSS-Eval Images metrics. The ``mir_eval`` module contains
    an implementation of BSS-Eval version 3.

    The BSS-Eval metrics attempt to measure perceptual quality by comparing sources
    estimated from a source separation algorithm to the ground truth, known sources.
    These metrics evaluate the distortion (SDR) and artifacts (SAR) present in the
    estimated signals as well as the interference (SIR) from other sources in a given
    estimated source. Results are returned in units of dB, with higher values indicating
    better quality.

    See Also:
        * For more information on ``mir_eval`` (python implementation of BSS-Eval v3) see
        `its Github page <https://github.com/craffel/mir_eval>`_.

        * For more information on the BSS-Eval metrics, see the webpage for
        `the original MATLAB implementation <http://bass-db.gforge.inria.fr/bss_eval/>`_.

        * Implementations of this base class: :class:`BSSEvalSources` and :class:`BSSEvalImages`.
        * :class:`BSSEvalV4` for the ``museval`` version 4 BSS-Eval implementation.

    References:
        * Emmanuel Vincent, Rémi Gribonval, Cédric Févotte. Performance measurement in blind
        audio source separation. IEEE Transactions on Audio, Speech and Language Processing,
        Institute of Electrical and Electronics Engineers, 2006, 14 (4), pp.1462–1469.
        <inria-00544230>
        * Colin Raffel, Brian McFee, Eric J. Humphrey, Justin Salamon, Oriol Nieto, Dawen Liang,
        and Daniel P. W. Ellis, "mir_eval: A Transparent Implementation of Common MIR Metrics",
        Proceedings of the 15th International Conference on Music Information Retrieval, 2014.

    Args:
        true_sources_list (list): List of :class:`AudioSignal` objects that contain the ground
            truth sources for the mixture.
        estimated_sources_list (list):  List of :class:`AudioSignal` objects that contain estimate
            sources, output from source separation algorithms.
        source_labels (list): List of strings that are labels for each source to be used as keys for
            the scores. Default value is ``None`` and in that case labels are ``Source 0``,
            ``Source 1``, etc.
        algorithm_name (str): Name of the algorithm if using this object to compute many
            BSS-Eval metrics. Can be changed later.
        do_mono (bool): Should flatten the audio to mono before calculating metrics.
        compute_permutation (bool): Should try to find the best permutation for the estimated
            sources.
    """

    def __init__(self, true_sources_list, estimated_sources_list, source_labels=None,
                 algorithm_name=None, do_mono=False, compute_permutation=True):
        super(BSSEvalImages, self).__init__(true_sources_list=true_sources_list,
                                            estimated_sources_list=estimated_sources_list,
                                            source_labels=source_labels, do_mono=do_mono,
                                            compute_permutation=compute_permutation)

        self._mir_eval_func = museval.metrics.bss_eval_images

    def _preprocess_sources(self):
        reference, estimated = super(BSSEvalImages, self)._preprocess_sources()

        if self.num_channels == 1:
            raise Exception("Can't run bss_eval_images on mono audio signals!")

        museval.metrics.validate(reference, estimated)

        return reference, estimated

    def _populate_scores_dict(self, bss_output):
        sdr_list, isr_list, sir_list, sar_list, perm = bss_output  # Unpack
        assert len(sdr_list) == len(sir_list) \
               == len(sar_list) == len(isr_list) == len(self.true_sources_list) * self.num_channels

        self.scores[self.RAW_VALUES] = {self.SDR: sdr_list, self.ISR: isr_list, self.SIR: sir_list,
                                        self.SAR: sar_list, self.PERMUTATION: perm}

        idx = 0
        for i, label in enumerate(self.source_labels):
            self.scores[label] = {}
            for ch in range(self.num_channels):
                chan = 'Ch {}'.format(ch)
                self.scores[label][chan] = {}

                self.scores[label][chan][self.SDR] = sdr_list[perm[idx]]
                self.scores[label][chan][self.ISR] = isr_list[perm[idx]]
                self.scores[label][chan][self.SIR] = sir_list[perm[idx]]
                self.scores[label][chan][self.SAR] = sar_list[perm[idx]]
                idx += 1

        self.scores[self.PERMUTATION] = perm



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Tests for PrecisionRecallFScore class

"""
import unittest
import os
import numpy as np
import sklearn.metrics

import nussl


class PrecisionRecallFScoreUnitTest(unittest.TestCase):

    def test_simple(self):
        drums_path = os.path.join('..', 'Input', 'src1.wav')
        flute_path = os.path.join('..', 'Input', 'src2.wav')

        drums = nussl.AudioSignal(drums_path)
        flute = nussl.AudioSignal(flute_path)
        flute.truncate_samples(drums.signal_length)

        mixture = drums + flute

        repet = nussl.Repet(mixture, mask_type=nussl.separation.BinaryMask)
        repet_mask_list = repet()

        ideal_mask = nussl.IdealMask(mixture, [drums, flute], mask_type=nussl.separation.BinaryMask)
        ideal_mask_list = ideal_mask()

        prf = nussl.PrecisionRecallFScore(ideal_mask_list, repet_mask_list)
        prf_scores = prf.evaluate()

    def test_prf_values(self):
        sizes = [128, 256, 512, 1024, 2048]
        for size in sizes:
            mask1_array = np.random.randint(0, 2, size=[size, size])
            mask1 = nussl.separation.BinaryMask(mask1_array)

            mask2_array = np.random.randint(0, 2, size=[size, size])
            mask2 = nussl.separation.BinaryMask(mask2_array)

            prf = nussl.PrecisionRecallFScore([mask1], [mask2])
            prf_scores = prf.evaluate()

            precision = sklearn.metrics.precision_score(mask1_array.ravel(), mask2_array.ravel())
            recall = sklearn.metrics.recall_score(mask1_array.ravel(), mask2_array.ravel())
            f1_score = sklearn.metrics.f1_score(mask1_array.ravel(), mask2_array.ravel())
            accuracy = sklearn.metrics.accuracy_score(mask1_array.ravel(), mask2_array.ravel())

            assert prf_scores['Source 0']['Precision'] == precision
            assert prf_scores['Source 0']['Recall'] == recall
            assert prf_scores['Source 0']['F1-Score'] == f1_score
            assert prf_scores['Source 0']['Accuracy'] == accuracy



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import sklearn

import separation_base
from ..core import constants
from ..core.audio_signal import AudioSignal
from ..core import utils


class ICA(separation_base.SeparationBase):
    """Separate sources using the Independent Component Analysis, given observations of the audio scene.
    nussl's ICA is a wrapper for sci-kit learn's implementation of FastICA, and provides a way to interop between
    nussl's :ref:`AudioSignal` objects and FastICA.

    References:
        `sci-kit learn FastICA <http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.fastica.html>`_

    Parameters:
        observations_list: multichannel AudioSignal object containing each observation of the mixture in each channel.
        Can prepare this input_audio_signal from multiple AudioSignal objects using
        ICA.numpy_observations_to_audio_signal(observations), where observations is a list of AudioSignal objects.

    """

    def __init__(self, observations_list=None, sample_rate=constants.DEFAULT_SAMPLE_RATE,
                 max_iterations=None, random_seed=None, fast_ica_kwargs=None):
        observations_signal = self._validate_observations_list(observations_list, sample_rate)
        super(ICA, self).__init__(input_audio_signal=observations_signal)

        # FastICA setup attributes
        self.num_components = self.audio_signal.num_channels

        self.fast_ica_kwargs = fast_ica_kwargs if isinstance(fast_ica_kwargs, dict) else {}
        self.max_iters = self._get_default_or_key(max_iterations, 'max_iter', self.fast_ica_kwargs)
        self.random_seed = self._get_default_or_key(random_seed, 'random_state', self.fast_ica_kwargs)

        # Results attributes
        self.estimated_sources = None
        self.estimated_mixing_params = None
        self.mean = None

    def _validate_observations_list(self, observations_list, sample_rate=None):
        """
        Validation for the observation list, can be a numpy array or list of AudioSignals with mono audio data
        Args:
            observations_list:

        Returns:

        """

        if isinstance(observations_list, np.ndarray):
            return self.numpy_observations_to_audio_signal(observations_list, sample_rate)

        elif isinstance(observations_list, list):
            return self.audio_signal_observations_to_audio_signal(observations_list)

        else:
            raise ValueError('Expected numpy array or list of AudioSignal objects,'
                             ' but got {}!'.format(type(observations_list)))

    @staticmethod
    def numpy_observations_to_audio_signal(observations, sample_rate=constants.DEFAULT_SAMPLE_RATE):
        """

        Args:
            observations (:obj:`np.ndarray`):
            sample_rate (int):

        Returns:

        """
        assert isinstance(observations, np.ndarray), 'Observations must be a numpy array!'
        if observations.ndim > 1\
                and observations.shape[constants.CHAN_INDEX] > observations.shape[constants.LEN_INDEX]:
            observations = observations.T

        return AudioSignal(audio_data_array=observations, sample_rate=sample_rate)

    @staticmethod
    def audio_signal_observations_to_audio_signal(observations):
        """

        Args:
            observations:

        Returns:

        """
        observations = utils.verify_audio_signal_list_strict(observations)

        if not all(observations[0].signal_length == o.signal_length for o in observations):
            raise ValueError('All observation AudioSignal objects must have the same length!')

        if not all(o.is_mono for o in observations):
            raise ValueError('All AudioSignals in observations_list must be mono!')

        observation_data = np.vstack([o.audio_data for o in observations])
        return AudioSignal(audio_data_array=observation_data, sample_rate=observations[0].sample_rate)

    @staticmethod
    def _get_default_or_key(default_value, key, dict_):
        if default_value is not None:
            return default_value
        elif key in dict_:
            return dict_[key]
        else:
            return None

    def run(self):
        """

        Returns:
            background (AudioSignal): An AudioSignal object with repeating background in background.audio_data
            (to get the corresponding non-repeating foreground run self.make_audio_signals())

        Example:
             ::

        """
        ica = sklearn.decomposition.FastICA(n_components=self.num_components, random_state=self.random_seed,
                                            max_iter=self.max_iters, **self.fast_ica_kwargs)

        # save for normalizing the estimated signals
        max_input_amplitude = np.max(np.abs(self.audio_signal.audio_data))

        # run ICA
        ica_output = ica.fit_transform(self.audio_signal.audio_data.T).T

        # now normalize the estimated signals
        max_output_amplitude = np.max(np.abs(ica_output))
        ica_output /= max_output_amplitude
        ica_output *= max_input_amplitude

        # store the resultant computations
        self.estimated_mixing_params = ica.mixing_
        self.mean = ica.mean_
        self.estimated_sources = [AudioSignal(audio_data_array=ica_output[i, :],
                                              sample_rate=self.audio_signal.sample_rate)
                                  for i in range(ica_output.shape[0])]

        return self.estimated_sources

    def make_audio_signals(self):
        """ Returns the background and foreground audio signals. You must have run ICA.run() prior
        to calling this function. This function will return None if run() has not been called.

        Returns:
            Audio Signals (list): list

        EXAMPLE:
             ::
        """
        if self.estimated_sources is None:
            raise ValueError('ICA.run() must be run prior to calling ICA.make_audio_signals()!')

        return self.estimated_sources



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The *nussl* External File Zoo (EFZ) is a server that houses all files that are too large to
bundle with *nussl* when distributing it through ``pip`` or Github. These types of files include
audio examples, benchmark files for tests, and trained neural network models.

*nussl* has built-in utilities for accessing the EFZ through its API. Here, it is possible to
see what files are available on the EFZ and download desired files. The EFZ utilities allow
for such functionality.
"""

import warnings
import json
import os
import sys
import hashlib

from six.moves.urllib_parse import urljoin
from six.moves.urllib.error import HTTPError
from six.moves.urllib.error import URLError
from six.moves.urllib.request import urlopen, Request

import constants


__all__ = ['get_available_audio_files', 'print_available_audio_files',
           'get_available_benchmark_files', 'print_available_benchmark_files',
           'get_available_trained_models', 'print_available_trained_models',
           'download_audio_file', 'download_benchmark_file', 'download_trained_model',
           'FailedDownloadError', 'MismatchedHashError', 'MetadataError']


def get_available_audio_files():
    """
    Returns a list of dicts containing metadata of the available audio files on the nussl External
    File Zoo (EFZ) server (http://nussl.ci.northwestern.edu/).

    Each entry in the list is in the following format:

    .. code-block:: python

        {
            u'file_length_seconds': 5.00390022675737,
            u'visible': True,
            u'file_name': u'K0140.wav',
            u'date_modified': u'2018-06-01',
            u'file_hash': u'f0d8d3c8d199d3790b0e42d1e5df50a6801f928d10f533149ed0babe61b5d7b5',
            u'file_size_bytes': 441388,
            u'file_description': u'Acoustic piano playing middle C.',
            u'audio_attributes': u'piano, middle C',
            u'file_size': u'431.0KiB',
            u'date_added': u'2018-06-01'
        }

    See Also:
        * :func:`print_available_audio_files`, prints a list of the audio files to the console.
        * :func:`download_audio_file` to download an audio file from the EFZ.

    Returns:
        (list): A list of dicts containing metadata of the available audio files on the nussl
        External File Zoo (EFZ) server (http://nussl.ci.northwestern.edu/).

    """
    # _download_all_metadata() will throw its own errors, so no try block around it
    return _download_all_metadata(constants.NUSSL_EFZ_AUDIO_METADATA_URL)


def print_available_audio_files():
    """
    Prints a message to the console that shows all of the available audio files that are on the
    nussl External File Zoo (EFZ) server (http://nussl.ci.northwestern.edu/).

    See Also:
        * :func:`get_available_audio_files` to get this same data from the EFZ server as a list.
        * :func:`download_audio_file` to download an audio file from the EFZ.

    Example:
        >>> import nussl
        >>> nussl.efz_utils.print_available_audio_files()
        File Name                                Duration (sec)  Size       Description
        dev1_female3_inst_mix.wav                10.0            1.7MiB     Instantaneous mixture of three female speakers talking in a stereo field.
        dev1_female3_synthconv_130ms_5cm_mix.wav 10.0            1.7MiB     Three female speakers talking in a stereo field, with 130ms of inter-channel delay.
        K0140.wav                                5.0             431.0KiB   Acoustic piano playing middle C.
        K0149.wav                                5.0             430.0KiB   Acoustic piano playing the A above middle C. (A440)

    To download one of these files insert the file name as the first parameter to
    :func:`download_audio_file`, like so:

    >>> nussl.efz_utils.download_audio_file('K0140.wav')

    """
    file_metadata = get_available_audio_files()

    print('{:40} {:15} {:10} {:50}'.format('File Name', 'Duration (sec)',
                                           'Size', 'Description'))
    for f in file_metadata:
        print('{:40} {:<15.1f} {:10} {:50}'.format(f['file_name'], f['file_length_seconds'],
                                                   f['file_size'], f['file_description']))
    print('To download one of these files insert the file name '
          'as the first parameter to nussl.download_audio_file(), like so: \n'
          ' >>> nussl.efz_utils.download_audio_file(\'K0140.wav\')')


def get_available_trained_models():
    """
    Returns a list of dicts containing metadata of the available trained models on the nussl
    External File Zoo (EFZ) server (http://nussl.ci.northwestern.edu/).

    Each entry in the list is in the following format:

    .. code-block:: python

        {
            u'for_class': u'DeepClustering',
            u'visible': True,
            u'file_name': u'deep_clustering_vocals_44k_long.model',
            u'date_modified': u'2018-06-01',
            u'file_hash': u'e09034c2cb43a293ece0b121f113b8e4e1c5a247331c71f40cb9ca38227ccc2c',
            u'file_size_bytes': 94543355,
            u'file_description': u'Deep clustering for vocal separation trained on augmented DSD100.',
            u'file_size': u'90.2MiB',
            u'date_added': u'2018-06-01'
        }

    Notes:
        Most of the entries in the dictionary are self-explanatory, but note the ``for_class``
        entry. The ``for_class`` entry specifies which `nussl` separation class the given model will
        work with. Usually, `nussl` separation classes that require a model will default so
        retrieving a model on the EFZ server (if not already found on the user's machine), but
        sometimes it is desirable to use a model other than the default one provided. In this case,
        the ``for_class`` entry lets the user know which class it is valid for use with.
        Additionally, trying to load a model into a class that it is not explicitly labeled for that
        class will raise an exception. Just don't do it, ok?

    See Also:
        * :func:`print_available_trained_models`, prints a list of the trained models to
            the console.
        * :func:`download_trained_model` to download a trained model from the EFZ.

    Returns:
        (list): A list of dicts containing metadata of the available trained models on the nussl
        External File Zoo (EFZ) server (http://nussl.ci.northwestern.edu/).
    """
    return _download_all_metadata(constants.NUSSL_EFZ_MODEL_METADATA_URL)


def print_available_trained_models():
    """
    Prints a message to the console that shows all of the available trained models that are on the
    nussl External File Zoo (EFZ) server (http://nussl.ci.northwestern.edu/).

    Notes:
        Most of the entries in the dictionary are self-explanatory, but note the ``for_class``
        entry. The ``for_class`` entry specifies which `nussl` separation class the given model will
        work with. Usually, `nussl` separation classes that require a model will default so
        retrieving a model on the EFZ server (if not already found on the user's machine), but
        sometimes it is desirable to use a model other than the default one provided. In this case,
        the ``for_class`` entry lets the user know which class it is valid for use with.
        Additionally, trying to load a model into a class that it is not explicitly labeled for that
        class will raise an exception. Just don't do it, ok?

    See Also:
        * :func:`get_available_trained_models` to get this same data from the EFZ server as a list.
        * :func:`download_trained_model` to download a trained model from the EFZ.

    Example:
        >>> import nussl
        >>> nussl.efz_utils.print_available_trained_models()
        File Name                                For Class            Size       Description
        deep_clustering_model.model              DeepClustering       48.1MiB    example Deep Clustering model
        deep_clustering_vocal_44k_long.model     DeepClustering       90.2MiB    trained DC model for vocal extraction

    To download one of these files insert the file name as the first parameter to download_trained_model(), like so:

    >>> nussl.efz_utils.download_trained_model('deep_clustering_model.h5')

    """
    file_metadata = get_available_trained_models()

    print('{:40} {:20} {:10} {:50}'.format('File Name', 'For Class', 'Size', 'Description'))
    for f in file_metadata:
        print('{:40} {:20} {:10} {:50}'.format(f['file_name'], f['for_class'],
                                               f['file_size'], f['file_description']))
    print('To download one of these files insert the file name '
          'as the first parameter to nussl.download_trained_model, like so: \n'
          ' >>> nussl.efz_utils.download_trained_model(\'deep_clustering_model.h5\')')


def get_available_benchmark_files():
    """
    Returns a list of dicts containing metadata of the available benchmark files for tests on the
    nussl External File Zoo (EFZ) server (http://nussl.ci.northwestern.edu/).

    Each entry in the list is in the following format:

    .. code-block:: python

        {
            u'for_class': u'DuetUnitTests',
            u'visible': True, u'file_name':
            u'benchmark_atn_bins.npy',
            u'date_modified': u'2018-06-19',
            u'file_hash': u'cf7fef6f4ea9af3dbde8b9880602eeaf72507b6c78f04097c5e79d34404a8a1f',
            u'file_size_bytes': 488,
            u'file_description': u'Attenuation bins numpy array for DUET benchmark test.',
            u'file_size': u'488.0B',
            u'date_added': u'2018-06-19'
        }

    Notes:
        Most of the entries in the dictionary are self-explanatory, but note the `for_class`
        entry. The `for_class` entry specifies which `nussl` benchmark class will load the
        corresponding benchmark file. Make sure these match exactly when writing tests!

    See Also:
        * :func:`print_available_benchmark_files`, prints a list of the benchmark files to the
            console.
        * :func:`download_benchmark_file` to download an benchmark file from the EFZ.

    Returns:
        (list): A list of dicts containing metadata of the available audio files on the nussl
        External File Zoo (EFZ) server (http://nussl.ci.northwestern.edu/).

    """
    return _download_all_metadata(constants.NUSSL_EFZ_BENCHMARK_METADATA_URL)


def print_available_benchmark_files():
    """
    Prints a message to the console that shows all of the available benchmark files that are on the
    nussl External File Zoo (EFZ) server (http://nussl.ci.northwestern.edu/).

    Example:
        >>> import nussl
        >>> nussl.efz_utils.print_available_benchmark_files()
        File Name                                For Class            Size       Description
        mix3_matlab_repet_foreground.mat         TestRepet            6.4MiB     Foreground matrix for Repet class benchmark test.
        benchmark_atn_bins.npy                   DuetUnitTests        488.0B     Attenuation bins numpy array for DUET benchmark test.
        benchmark_sym_atn.npy                    DuetUnitTests        3.4MiB     Symmetric attenuation histogram for the DUET benchmark test.
        benchmark_wmat.npy                       DuetUnitTests        3.4MiB     Frequency matrix for the DUET benchmark test.

    To download one of these files insert the file name as the first parameter to nussl.download_benchmark_file, like so:

    >>> nussl.efz_utils.download_benchmark_file('example.npy')

    Notes:
        Most of the entries in the printed list are self-explanatory, but note the ``for_class``
        entry. The ``for_class`` entry specifies which `nussl` benchmark class will load the
        corresponding benchmark file. Make sure these match exactly when writing tests!

    See Also:
        * :func:`get_available_benchmark_files`, prints a list of the benchmark files to the
            console.
        * :func:`download_benchmark_file` to download an benchmark file from the EFZ.

    """
    file_metadata = get_available_benchmark_files()

    print('{:40} {:20} {:10} {:50}'.format('File Name', 'For Class', 'Size', 'Description'))
    for f in file_metadata:
        print('{:40} {:20} {:10} {:50}'.format(f['file_name'], f['for_class'],
                                               f['file_size'], f['file_description']))
    print('To download one of these files insert the file name '
          'as the first parameter to nussl.download_benchmark_file, like so: \n'
          ' >>> nussl.efz_utils.download_benchmark_file(\'example.npy\')')


def _download_all_metadata(url):
    """
    Downloads the json file that contains all of the metadata for a specific file type (read:
    audio files, benchmark files, or trained models) that is on the EFZ server. This is retrieved
    from one of following three URLs (which are stored in nussl.constants):
    NUSSL_EFZ_AUDIO_METADATA_URL, NUSSL_EFZ_BENCHMARK_METADATA_URL, or NUSSL_EFZ_MODEL_METADATA_URL.

    Args:
        url (str):  URL for the EFZ server that has metadata. One of these three:
            NUSSL_EFZ_AUDIO_METADATA_URL, NUSSL_EFZ_BENCHMARK_METADATA_URL, or
            NUSSL_EFZ_MODEL_METADATA_URL.

    Returns:
        (list): List of dicts with metadata for the desired file type.

    """
    request = Request(url)

    # Make sure to get the newest data
    request.add_header('Pragma', 'no-cache')
    request.add_header('Cache-Control', 'max-age=0')
    response = urlopen(request)
    return json.loads(response.read())


def _download_metadata_for_file(file_name, file_type):
    """
    Downloads the metadata entry for a specific file (:param:`file_name`) on the EFZ server.

    Args:
        file_name (str): File name as specified on the EFZ server.
        file_type (str): 'Type' of file, either 'audio', 'model', or 'benchmark'.

    Returns:
        (dict) Metadata entry for the specified file, or ``None`` if it could not be located.

    """

    metadata_urls = {
        'audio': constants.NUSSL_EFZ_AUDIO_METADATA_URL,
        'benchmark': constants.NUSSL_EFZ_BENCHMARK_METADATA_URL,
        'model': constants.NUSSL_EFZ_MODEL_METADATA_URL,
    }

    if metadata_urls[file_type]:
        metadata_url = metadata_urls[file_type]
    else:
        # wrong file type, return
        raise MetadataError('Cannot find metadata of type {}.'.format(file_type))

    metadata = _download_all_metadata(metadata_url)

    for file_metadata in metadata:
        if file_metadata['file_name'] == file_name:
            return file_metadata

    raise MetadataError('No matching metadata for file {} at url {}!'
                        .format(file_name, constants.NUSSL_EFZ_AUDIO_METADATA_URL))


def download_audio_file(audio_file_name, local_folder=None, verbose=True):
    """
    Downloads the specified audio file from the `nussl` External File Zoo (EFZ) server. The
    downloaded file is stored in :param:`local_folder` if a folder is provided. If a folder is
    not provided, `nussl` attempts to save the downloaded file in `~/.nussl/` (expanded) or in
    `tmp/.nussl`. If the requested file is already in :param:`local_folder` (or one of the two
    aforementioned directories) and the calculated hash matches the precomputed hash from the EFZ
    server metadata, then the file will not be downloaded.

    Args:
        audio_file_name: (str) Name of the audio file to attempt to download.
        local_folder: (str) Path to local folder in which to download the file.
            If no folder is provided, `nussl` will store the file in `~/.nussl/` (expanded) or in
            `tmp/.nussl`.
        verbose (bool): If ``True`` prints the status of the download to the console.

    Returns:
        (String) Full path to the requested file (whether downloaded or not).

    Example:
        >>> import nussl
        >>> piano_path = nussl.efz_utils.download_audio_file('K0140.wav')
        >>> piano_signal = nussl.AudioSignal(piano_path)

    """
    file_metadata = _download_metadata_for_file(audio_file_name, 'audio')

    file_hash = file_metadata['file_hash']

    file_url = urljoin(constants.NUSSL_EFZ_AUDIO_URL, audio_file_name)
    result = _download_file(audio_file_name, file_url, local_folder, 'audio',
                            file_hash=file_hash, verbose=verbose)

    return result


def download_trained_model(model_name, local_folder=None, verbose=True):
    """
    Downloads the specified trained model from the `nussl` External File Zoo (EFZ) server. The
    downloaded file is stored in :param:`local_folder` if a folder is provided. If a folder is
    not provided, `nussl` attempts to save the downloaded file in `~/.nussl/` (expanded) or in
    `tmp/.nussl`. If the requested file is already in :param:`local_folder` (or one of the two
    aforementioned directories) and the calculated hash matches the precomputed hash from the EFZ
    server metadata, then the file will not be downloaded.

    Args:
        audio_file_name: (str) Name of the trained model to attempt to download.
        local_folder: (str) Path to local folder in which to download the file.
            If no folder is provided, `nussl` will store the file in `~/.nussl/` (expanded) or in
            `tmp/.nussl`.
        verbose (bool): If ``True`` prints the status of the download to the console.

    Returns:
        (String) Full path to the requested file (whether downloaded or not).

    Example:
        >>> import nussl
        >>> model_path = nussl.efz_utils.download_trained_model('deep_clustering_model.h5')
        >>> signal = nussl.AudioSignal()
        >>> piano_signal = nussl.DeepClustering(signal, model_path=model_path)

    """
    file_metadata = _download_metadata_for_file(model_name, 'model')

    file_hash = file_metadata['file_hash']

    file_url = urljoin(constants.NUSSL_EFZ_MODELS_URL, model_name)
    result = _download_file(model_name, file_url, local_folder, 'models',
                            file_hash=file_hash, verbose=verbose)

    return result


def download_benchmark_file(benchmark_name, local_folder=None, verbose=True):
    """
    Downloads the specified benchmark file from the `nussl` External File Zoo (EFZ) server. The
    downloaded file is stored in :param:`local_folder` if a folder is provided. If a folder is
    not provided, `nussl` attempts to save the downloaded file in `~/.nussl/` (expanded) or in
    `tmp/.nussl`. If the requested file is already in :param:`local_folder` (or one of the two
    aforementioned directories) and the calculated hash matches the precomputed hash from the EFZ
    server metadata, then the file will not be downloaded.

    Args:
        audio_file_name: (str) Name of the trained model to attempt to download.
        local_folder: (str) Path to local folder in which to download the file.
            If no folder is provided, `nussl` will store the file in `~/.nussl/` (expanded) or in
            `tmp/.nussl`.
        verbose (bool): If ``True`` prints the status of the download to the console.

    Returns:
        (String) Full path to the requested file (whether downloaded or not).

    Example:
        >>> import nussl
        >>> import numpy as np
        >>> stm_atn_path = nussl.efz_utils.download_benchmark_file('benchmark_sym_atn.npy')
        >>> sym_atm = np.load(stm_atn_path)

    """
    file_metadata = _download_metadata_for_file(benchmark_name, 'benchmark')

    file_hash = file_metadata['file_hash']

    file_url = urljoin(constants.NUSSL_EFZ_BENCHMARKS_URL, benchmark_name)
    result = _download_file(benchmark_name, file_url, local_folder, 'benchmarks',
                            file_hash=file_hash, verbose=verbose)

    return result


def _download_file(file_name, url, local_folder, cache_subdir,
                   file_hash=None, cache_dir=None, verbose=True):
    """
    Downloads the specified file from the

    Heavily inspired by and lovingly adapted from keras' `get_file` function:
    https://github.com/fchollet/keras/blob/afbd5d34a3bdbb0916d558f96af197af1e92ce70/keras/utils/data_utils.py#L109

    Args:
        file_name: (String) name of the file located on the server
        url: (String) url of the file
        local_folder: (String) alternate folder in which to download the file
        cache_subdir: (String) subdirectory of folder in which to download flie
        file_hash: (String) expected hash of downloaded file
        cache_dir:

    Returns:
        (String) local path to downloaded file

    """
    if local_folder not in [None, '']:
        # local folder provided, let's create it if it doesn't exist and use it as datadir
        if not os.path.exists(os.path.expanduser(local_folder)):
            os.makedirs(os.path.expanduser(local_folder))
        datadir = os.path.expanduser(local_folder)

    else:
        if cache_dir is None:
            cache_dir = os.path.expanduser(os.path.join('~', '.nussl'))
        datadir_base = os.path.expanduser(cache_dir)

        if not os.access(datadir_base, os.W_OK):
            datadir_base = os.path.join('/tmp', '.nussl')

        datadir = os.path.join(datadir_base, cache_subdir)
        if not os.path.exists(datadir):
            os.makedirs(datadir)

    file_path = os.path.join(datadir, file_name)

    download = False
    if os.path.exists(file_path):
        if file_hash is not None:
            # compare the provided hash with the hash of the file currently at file_path
            current_hash = _hash_file(file_path)

            # if the hashes are equal, we already have the file we need, so don't download
            if file_hash != current_hash:
                if verbose:
                    warnings.warn("Hash for {} does not match known hash. "
                                  "Downloading {} from servers...".format(file_path, file_name))
                download = True
            elif verbose:
                print('Matching file found at {}, skipping download.'.format(file_path))

        else:
            download = True

    else:
        download = True

    if download:
        if verbose:
            print('Saving file at {}'.format(file_path))
            print('Downloading {} from {}'.format(file_name, url))

        def _dl_progress(count, block_size, total_size):
            percent = int(count * block_size * 100 / total_size)

            if percent <= 100:
                sys.stdout.write('\r{}...{}%'.format(file_name, percent))
                sys.stdout.flush()

        error_msg = 'URL fetch failure on {}: {} -- {}'

        try:
            try:
                reporthook = _dl_progress if verbose else None
                urlretrieve(url, file_path, reporthook)
                if verbose: print()  # print a new line after the progress is done.

            except HTTPError as e:
                raise FailedDownloadError(error_msg.format(url, e.code, e.msg))
            except URLError as e:
                raise FailedDownloadError(error_msg.format(url, e.errno, e.reason))
        except (Exception, KeyboardInterrupt) as e:
            if os.path.exists(file_path):
                os.remove(file_path)
            raise e

        # check hash of received file to see if it matches the provided hash
        if file_hash is not None:
            download_hash = _hash_file(file_path)
            if file_hash != download_hash:
                # the downloaded file is not what it should be. Get rid of it.
                os.remove(file_path)
                raise MismatchedHashError("Downloaded file ({}) has been deleted "
                                          "because of a hash mismatch.".format(file_path))

        return file_path

    else:
        return file_path


def _hash_file(file_path, chunk_size=65535):
    """

    Args:
        file_path: System path to the file to be hashed
        chunk_size: size of chunks

    Returns:
        file_hash: the SHA256 hashed string in hex

    """
    hasher = hashlib.sha256()

    with open(file_path, 'rb') as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b''):
            hasher.update(chunk)

    return hasher.hexdigest()


# This is stolen wholesale from keras
if sys.version_info[0] == 2:
    def urlretrieve(url, filename, reporthook=None, data=None):
        """Replacement for `urlretrive` for Python 2.
        Under Python 2, `urlretrieve` relies on `FancyURLopener` from legacy
        `urllib` module, known to have issues with proxy management.
        # Arguments
            url: url to retrieve.
            filename: where to store the retrieved data locally.
            reporthook: a hook function that will be called once
                on establishment of the network connection and once
                after each block read thereafter.
                The hook will be passed three arguments;
                a count of blocks transferred so far,
                a block size in bytes, and the total size of the file.
            data: `data` argument passed to `urlopen`.
        """
        def chunk_read(response, chunk_size=8192, reporthook=None):
            content_type = response.info().get('Content-Length')
            total_size = -1
            if content_type is not None:
                total_size = int(content_type.strip())
            count = 0
            while 1:
                chunk = response.read(chunk_size)
                count += 1
                if not chunk:
                    reporthook(count, total_size, total_size)
                    break
                if reporthook:
                    reporthook(count, chunk_size, total_size)
                yield chunk

        response = urlopen(url, data)
        with open(filename, 'wb') as fd:
            for chunk in chunk_read(response, reporthook=reporthook):
                fd.write(chunk)
else:
    from six.moves.urllib.request import urlretrieve  # pylint: disable=g-import-not-at-top


########################################
#             Error Classes
########################################


class FailedDownloadError(Exception):
    """
    Exception class for failed file downloads.
    """
    pass


class MismatchedHashError(Exception):
    """
    Exception class for when a computed hash function does match a pre-computed hash.
    """
    pass


class MetadataError(Exception):
    """
    Exception class for errors with metadata.
    """
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .. import torch_imported, vamp_imported, ImportErrorClass

# Import base classes
from .masks import *
from .separation_base import SeparationBase
from .mask_separation_base import MaskSeparationBase

# Median based algorithms
from .repet import Repet
from .repet_sim import RepetSim
from .ft2d import FT2D
from .hpss import HPSS

median_algorithms = [Repet, RepetSim, FT2D, HPSS]

# Melody-based methods
if vamp_imported:
    from .melodia import Melodia
else:
    class Melodia(ImportErrorClass):
        def __init__(self, **kwargs):
            super(Melodia, self).__init__('vamp')

melody_algorithms = [Melodia]

# Spatialization based methods
from .duet import Duet
from .projet import Projet

spatialization_algorithms = [Duet, Projet]

# Benchmark algorithms
from .ideal_mask import IdealMask
from .high_low_pass_filter import HighLowPassFilter

benchmark_algorithms = [IdealMask, HighLowPassFilter]

# Composite algorithms
from .overlap_add import OverlapAdd
composite_instruments = [OverlapAdd]

# Matrix factorization and component analysis
from .nmf_mfcc import NMF_MFCC
from .ica import ICA
from .rpca import RPCA

nmf_algorithms = [NMF_MFCC]
component_analysis_algorithms = [ICA, RPCA]

# Deep learning algorithms
if torch_imported:
    from .deep_clustering import DeepClustering
else:
    class DeepClustering(ImportErrorClass):
        def __init__(self, **kwargs):
            super(DeepClustering, self).__init__('pytorch')

deep_learning_algorithms = [DeepClustering]

all_separation_algorithms = [median_algorithms, melody_algorithms, spatialization_algorithms,
                             benchmark_algorithms, nmf_algorithms, component_analysis_algorithms,
                             deep_learning_algorithms]

all_separation_algorithms = [val for sublist in all_separation_algorithms for val in sublist]

__all__ = ['SeparationBase', 'MaskSeparationBase',
           'all_separation_algorithms',
           'median_algorithms', 'Repet', 'RepetSim', 'HPSS', 'FT2D',
           'melody_algorithms', 'Melodia',
           'spatialization_algorithms', 'Duet', 'Projet',
           'benchmark_algorithms', 'IdealMask', 'HighLowPassFilter',
           'composite_instruments', 'OverlapAdd',
           'nmf_algorithms', 'NMF_MFCC',
           'component_analysis_algorithms', 'ICA', 'RPCA',
           'deep_learning_algorithms', 'DeepClustering']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
An implementation of the Degenerate Unmixing Estimation Technique (DUET) algorithm.
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy import signal

import mask_separation_base
import masks
from ..core import utils
from ..core import constants
from ..core.audio_signal import AudioSignal


class Duet(mask_separation_base.MaskSeparationBase):
    """
    The DUET algorithm was originally proposed by S.Rickard and F.Dietrich for DOA estimation
    and further developed for BSS and demixing by A.Jourjine, S.Rickard, and O. Yilmaz. DUET extracts
    sources using the symmetric attenuation and relative delay between two channels. The symmetric
    attenuation is calculated from the ratio of the two channels' stft amplitudes, and the delay is the arrival delay
    between the two sensors used to record the audio signal. These two values are clustered as peaks on a
    histogram to determine where each source occurs. This implementation of DUET creates and returns Mask objects after
    the run() function, which can then be applied to the original audio signal to extract each individual source.

    References:

        * Rickard, Scott. "The DUET blind source separation algorithm." Blind Speech Separation. Springer Netherlands,
          2007. 217-241.
        * Yilmaz, Ozgur, and Scott Rickard. "Blind separation of speech mixtures via time-frequency masking." Signal
          Processing, IEEE transactions on 52.7 (2004): 1830-1847.

    Parameters:
        input_audio_signal (np.array): a 2-row Numpy matrix containing samples of the two-channel mixture.
        num_sources (int): Number of sources to find.
        attenuation_min (int): Minimum distance in utils.find_peak_indices, change if not enough peaks are identified.
        attenuation_max (int): Used for creating a histogram without outliers.
        num_attenuation_bins (int): Number of bins for attenuation. 
        delay_min (int): Lower bound on delay, used as minimum distance in utils.find_peak_indices.
        delay_max (int): Upper bound on delay, used for creating a histogram without outliers.
        num_delay_bins (int): Number of bins for delay. 
        peak_threshold (float): Value in [0, 1] for peak picking. 
        attenuation_min_distance (int): Minimum distance between peaks wrt attenuation. 
        delay_min_distance (int): Minimum distance between peaks wrt delay.
        p (int): Weight the histogram with the symmetric attenuation estimator.
        q (int): Weight the histogram with the delay estimator
        On page 8 of his paper, Rickard recommends p=1 and q=0 as a default starting point and p=.5, q=0 if one
        source is more dominant.

    Attributes:
        stft_ch0 (np.array): A Numpy matrix containing the stft data of channel 0.
        stft_ch1 (np.array): A Numpy matrix containing the stft data of channel 1.
        frequency_matrix (np.array): A Numpy matrix containing the frequencies of analysis.
        symmetric_atn (np.array): A Numpy matrix containing the symmetric attenuation between the two channels.
        delay (np.array): A Numpy matrix containing the delay between the two channels.
        num_time_bins (np.array): The number of time bins for the frequency matrix and mask arrays.
        num_frequency_bins (int): The number of frequency bins for the mask arrays.
        attenuation_bins (int): A Numpy array containing the attenuation bins for the histogram.
        delay_bins (np.array): A Numpy array containing the delay bins for the histogram.
        normalized_attenuation_delay_histogram (np.array): A normalized Numpy matrix containing the
                                                           attenuation delay histogram, which has peaks for each source.
        attenuation_delay_histogram (np.array): A non-normalized Numpy matrix containing the
                                                attenuation delay histogram, which has peaks for each source.
        peak_indices (np.array): A Numpy array containing the indices of the peaks for the histogram.
        separated_sources (np.array): A Numpy array of arrays containing each separated source.

    Examples:
        :ref:`The DUET Demo Example <duet_demo>`

    """

    def __init__(self, input_audio_signal, num_sources,
                 attenuation_min=-3, attenuation_max=3, num_attenuation_bins=50,
                 delay_min=-3, delay_max=3, num_delay_bins=50,
                 peak_threshold=0.2, attenuation_min_distance=5, delay_min_distance=5, p=1, q=0):
        super(Duet, self).__init__(input_audio_signal=input_audio_signal,
                                   mask_type=mask_separation_base.MaskSeparationBase.BINARY_MASK)

        if not self.audio_signal.is_stereo:
            raise ValueError('Duet requires that the input_audio_signal has exactly 2 channels!')

        self.num_sources = num_sources
        self.attenuation_min = attenuation_min
        self.attenuation_max = attenuation_max
        self.num_attenuation_bins = num_attenuation_bins
        self.delay_min = delay_min
        self.delay_max = delay_max
        self.num_delay_bins = num_delay_bins
        self.peak_threshold = peak_threshold
        self.attenuation_min_distance = attenuation_min_distance
        self.delay_min_distance = delay_min_distance
        self.p = p
        self.q = q
        self.final_sources = []
        self.result_masks = []

        self.stft_ch0 = None
        self.stft_ch1 = None
        self.frequency_matrix = None
        self.symmetric_atn = None
        self.delay = None
        self.num_time_bins = None
        self.num_frequency_bins = None
        self.attenuation_bins = None
        self.delay_bins = None
        self.normalized_attenuation_delay_histogram = None
        self.attenuation_delay_histogram = None
        self.peak_indices = None
        self.delay_peak = None
        self.atn_peak = None
        self.separated_sources = None

    def run(self):
        """ Extracts N sources from a given stereo audio mixture (N sources captured via 2 sensors)

        Returns:
            computed_masks (np.array): A list of binary mask objects that can be used to extract the sources

        Example:
        .. code-block:: python
            :linenos:

            #Import input audio signal
            input_file_name = '../Input/dev1_female3_inst_mix.wav'
            signal = AudioSignal(path_to_input_file=input_file_name)

            # Set up and run Duet
            duet = Duet(signal, a_min=-3, a_max=3, a_num=50, d_min=-3, d_max=3, d_num=50, threshold=0.2,
            a_min_distance=5, d_min_distance=5, num_sources=3)
            duet.run()

            # plot histogram results
            duet.plot(os.path.join('..', 'Output', 'duet_2d.png'))
            duet.plot(os.path.join('..', 'Output', 'duet_3d.png'), three_d_plot=True)

            # Create output file for each source found
            output_name_stem = os.path.join('..', 'Output', 'duet_source')
            i = 1
            for s in duet.make_audio_signals():
                output_file_name = output_name_stem + str(i) + '.wav'
                s.write_audio_to_file(output_file_name)
                i += 1

        """
        self.result_masks = []

        if not self.audio_signal.is_stereo:  # double check this
            raise ValueError('Cannot run Duet on audio signal without exactly 2 channels!')

        # Calculate the stft of both channels and create the frequency matrix (the matrix containing the
        #  frequencies of analysis of the Fourier transform)
        self.stft_ch0, self.stft_ch1, self.frequency_matrix = self._compute_spectrogram(self.sample_rate)

        # Calculate the symmetric attenuation (alpha) and delay (delta) for each
        # time-freq. point and return a matrix for each
        self.symmetric_atn, self.delay = self._compute_atn_delay(self.stft_ch0, self.stft_ch1, self.frequency_matrix)

        # Make histogram of attenuation-delay values and get the center values for the bins in this histogram
        self.normalized_attenuation_delay_histogram, self.attenuation_bins, self.delay_bins = self._make_histogram()

        # Find the location of peaks in the attenuation-delay plane
        self.peak_indices = utils.find_peak_indices(self.normalized_attenuation_delay_histogram, self.num_sources,
                                                    threshold=self.peak_threshold,
                                                    min_dist=[self.attenuation_min_distance,
                                                              self.delay_min_distance])

        # compute delay_peak, attenuation peak, and attenuation/delay estimates
        self.delay_peak, atn_delay_est, self.atn_peak = self._convert_peaks(self.peak_indices)

        # compute masks for separation
        computed_masks = self._compute_masks()

        return computed_masks

    def get_atn_delay_histogram(self, recompute=False, normalized=False):
        """
        Computes the 2D attenuation/delay histogram for the given input signal.
        Args:
            recompute (bool): If `True`, this will recompute everything required to calculate the histogram
            (i.e., spectrograms). If `False`, then this will use previously calculated values if available.
            normalized (bool): Will normalize the histogram

        Returns:

        """

        if not self.audio_signal.is_stereo:  # double check this
            raise ValueError('Cannot run Duet on audio signal without exactly 2 channels!')

        if not np.all([self.stft_ch0, self.stft_ch1, self.frequency_matrix]) or recompute:
            self.stft_ch0, self.stft_ch1, self.frequency_matrix = self._compute_spectrogram(self.sample_rate)

        if not np.all([self.symmetric_atn, self.delay]) or recompute:
            self.symmetric_atn, self.delay = self._compute_atn_delay(self.stft_ch0, self.stft_ch1,
                                                                     self.frequency_matrix)

        if self.normalized_attenuation_delay_histogram is None or recompute:
            self.normalized_attenuation_delay_histogram, self.attenuation_bins, self.delay_bins = self._make_histogram()

        if normalized:
            return self.normalized_attenuation_delay_histogram
        else:
            return self.attenuation_delay_histogram

    def convert_peaks_to_masks(self, peak_indices=None):
        """

        Args:
            peak_indices:

        Returns:

        """
        if peak_indices is None and self.peak_indices is None:
            raise ValueError('Need peak_indices to make masks!')

        if self.normalized_attenuation_delay_histogram is None:
            self.get_atn_delay_histogram(recompute=True)

        peak_indices = self.peak_indices if peak_indices is None else peak_indices
        self.delay_peak, atn_delay_est, self.atn_peak = self._convert_peaks(peak_indices)
        return self._compute_masks()

    def _compute_spectrogram(self, sample_rate):
        """ Creates the STFT matrices for channel 0 and 1, and computes the frequency matrix.
        Parameter:
            sample_rate (integer): sample rate

        Returns:
            stft_ch0 (np.matrix): a 2D Numpy matrix containing the stft of channel 0
            stft_ch1 (np.matrix): a 2D Numpy matrix containing the stft of channel 1
            wmat (np.matrix): a 2D Numpy matrix containing the frequencies of analysis of the Fourier transform
        """

        # Compute the stft of the two channel mixtures
        self.audio_signal.stft_params = self.stft_params
        self.audio_signal.stft()

        stft_ch0 = self.audio_signal.get_stft_channel(0)
        stft_ch1 = self.audio_signal.get_stft_channel(1)

        # Compute the freq. matrix for later use in phase calculations
        n_time_bins = len(self.audio_signal.time_bins_vector)
        wmat = np.array(np.tile(np.mat(self.audio_signal.freq_vector).T, (1, n_time_bins))) * (2 * np.pi / sample_rate)
        wmat += constants.EPSILON
        return stft_ch0, stft_ch1, wmat

    @staticmethod
    def _compute_atn_delay(stft_ch0, stft_ch1, frequency_matrix):
        # Calculate the symmetric attenuation (alpha) and delay (delta) for each
        # time-freq. point
        inter_channel_ratio = (stft_ch1 + constants.EPSILON) / (stft_ch0 + constants.EPSILON)
        attenuation = np.abs(inter_channel_ratio)  # relative attenuation between the two channels
        symmetric_attenuation = attenuation - 1 / attenuation  # symmetric attenuation
        relative_delay = -np.imag(np.log(inter_channel_ratio)) / (2 * np.pi * frequency_matrix)  # relative delay
        return symmetric_attenuation, relative_delay

    def _make_histogram(self):
        """Receives the stft of the two channel mixtures and the frequency matrix to a create
        a smooth and normalized histogram.

        Parameters:
        stft_ch0 (complex np.array): a 2D Numpy matrix containing the stft of channel 0
        stft_ch1 (complex np.array): a 2D Numpy matrix containing the stft of channel 1
        symmetric_atn (np.array): the symmetric attenuation between two channels
        delay (np.array): the time delay between 2 channels
        wmat(np.array): a 2D Numpy matrix containing the frequency matrix of the signal

        Returns:
            histogram (np.array): a smooth and normalized histogram
            atn_bins (np.array): The range of attenuation values distributed into bins
            delay_bins (np.array): The range of delay values distributed into bins
        """
        # calculate the weighted histogram
        time_frequency_weights = (np.abs(self.stft_ch0) * np.abs(self.stft_ch1)) ** self.p * \
                                 (np.abs(self.frequency_matrix)) ** self.q

        # only consider time-freq. points yielding estimates in bounds
        attenuation_premask = np.logical_and(self.attenuation_min < self.symmetric_atn,
                                             self.symmetric_atn < self.attenuation_max)

        delay_premask = np.logical_and(self.delay_min < self.delay, self.delay < self.delay_max)
        attenuation_delay_premask = np.logical_and(attenuation_premask, delay_premask)

        nonzero_premask = np.nonzero(attenuation_delay_premask)
        symmetric_attenuation_vector = self.symmetric_atn[nonzero_premask]
        delay_vector = self.delay[nonzero_premask]
        time_frequency_weights_vector = time_frequency_weights[nonzero_premask]

        bins_array = np.array([self.num_attenuation_bins, self.num_delay_bins])
        range_array = np.array([[self.attenuation_min, self.attenuation_max], [self.delay_min, self.delay_max]])

        # compute the histogram
        histogram, atn_bins, delay_bins = np.histogram2d(symmetric_attenuation_vector, delay_vector,
                                                         bins=bins_array, range=range_array,
                                                         weights=time_frequency_weights_vector)

        # Save non-normalized as an option for plotting later
        self.attenuation_delay_histogram = histogram

        # Scale histogram from 0 to 1
        histogram /= histogram.max()

        # smooth the normalized histogram - local average 3-by-3 neighboring bins
        histogram = self._smooth_matrix(histogram, np.array([3]))
        return histogram, atn_bins, delay_bins

    def _convert_peaks(self, peak_indices):
        """Receives the attenuation and delay bins and computes the delay/attenuation
        peaks based on the peak finder indices.

        Returns:
            delay_peak(np.array): The delay peaks determined from the histogram
            atn_delay_est (np.array): The estimated symmetric attenuation and delay values
            atn_peak (np.array): Attenuation converted from symmetric attenuation
        """

        if peak_indices is None:
            raise ValueError('No peak indices')

        atn_indices = [x[0] for x in peak_indices]
        delay_indices = [x[1] for x in peak_indices]

        if any(a > len(self.attenuation_bins) for a in atn_indices):
            raise ValueError('Attenuation index greater than length of bins')

        symmetric_atn_peak = self.attenuation_bins[atn_indices]
        delay_peak = self.delay_bins[delay_indices]

        atn_delay_est = np.column_stack((symmetric_atn_peak, delay_peak))

        # convert symmetric_atn to atn_peak using formula from Rickard
        atn_peak = (symmetric_atn_peak + np.sqrt(symmetric_atn_peak ** 2 + 4)) / 2
        return delay_peak, atn_delay_est, atn_peak

    def _compute_masks(self):
        """Receives the attenuation and delay peaks and computes a mask to be applied to the signal for source
        separation.

        """
        # compute masks for separation
        best_so_far = np.inf * np.ones_like(self.stft_ch0, dtype=float)

        for i in range(0, self.num_sources):
            mask_array = np.zeros_like(self.stft_ch0, dtype=bool)
            phase = np.exp(-1j * self.frequency_matrix * self.delay_peak[i])
            score = np.abs(self.atn_peak[i] * phase * self.stft_ch0 - self.stft_ch1) ** 2 / (1 + self.atn_peak[i] ** 2)
            mask = (score < best_so_far)
            mask_array[mask] = True
            background_mask = masks.BinaryMask(np.array(mask_array))
            self.result_masks.append(background_mask)
            self.result_masks[0].mask = np.logical_xor(self.result_masks[i].mask, self.result_masks[0].mask)
            best_so_far[mask] = score[mask]

        # Compute first mask based on what the other masks left remaining
        self.result_masks[0].mask = np.logical_not(self.result_masks[0].mask)
        return self.result_masks

    @staticmethod
    def _smooth_matrix(matrix, kernel):
        """Performs two-dimensional convolution in order to smooth the values of matrix elements.

        (similar to low-pass filtering)

        Parameters:
            matrix (np.array): a 2D Numpy matrix to be smoothed
            kernel (np.array): a 2D Numpy matrix containing kernel values
        Note:
            if Kernel is of size 1 by 1 (scalar), a Kernel by Kernel matrix of 1/Kernel**2 will be used as the matrix
            averaging kernel
        Output:
            smoothed_matrix (np.array): a 2D Numpy matrix containing a smoothed version of Mat (same size as Mat)
        """

        # check the dimensions of the Kernel matrix and set the values of the averaging
        # matrix, kernel_matrix
        if np.prod(kernel.shape) == 1:
            kernel_matrix = np.ones((kernel[0], kernel[0])) / kernel[0] ** 2
        else:
            kernel_matrix = kernel

        # make kernel_matrix have odd dimensions
        krow, kcol = np.shape(kernel_matrix)
        if np.mod(krow, 2) == 0:
            kernel_matrix = signal.convolve2d(kernel_matrix, np.ones((2, 1))) / 2
            krow += 1

        if np.mod(kcol, 2) == 0:
            kernel_matrix = signal.convolve2d(kernel_matrix, np.ones((1, 2))) / 2
            kcol += 1

        # adjust the matrix dimension for convolution
        copy_row = int(np.floor(krow / 2))  # number of rows to copy on top and bottom
        copy_col = int(np.floor(kcol / 2))  # number of columns to copy on either side

        # TODO: This is very ugly. Make this readable.
        # form the augmented matrix (rows and columns added to top, bottom, and sides)
        matrix = np.mat(matrix)  # make sure Mat is a Numpy matrix
        augmented_matrix = np.vstack(
            [
                np.hstack(
                    [matrix[0, 0] * np.ones((copy_row, copy_col)),
                     np.ones((copy_row, 1)) * matrix[0, :],
                     matrix[0, -1] * np.ones((copy_row, copy_col))
                     ]),
                np.hstack(
                    [matrix[:, 0] * np.ones((1, copy_col)),
                     matrix,
                     matrix[:, -1] * np.ones((1, copy_col))]),
                np.hstack(
                    [matrix[-1, 1] * np.ones((copy_row, copy_col)),
                     np.ones((copy_row, 1)) * matrix[-1, :],
                     matrix[-1, -1] * np.ones((copy_row, copy_col))
                     ]
                )
            ]
        )

        # perform two-dimensional convolution between the input matrix and the kernel
        smooted_matrix = signal.convolve2d(augmented_matrix, kernel_matrix[::-1, ::-1], mode='valid')

        return smooted_matrix

    def make_audio_signals(self):
        """Returns the extracted signals

        Returns:
            signals (List[AudioSignal]): List of AudioSignals extracted using DUET.

        """
        if len(self.result_masks) == 0:
            raise ValueError('Cannot make audio signals with no masks to apply!')

        signals = []
        for i in range(self.num_sources):
            # Apply masks to stft channels using equation 8.34 (pg. 11) provided by Rickard
            stft_sources = ((self.stft_ch0 + self.atn_peak[i] * np.exp(1j * self.frequency_matrix * self.delay_peak[i])
                             * self.stft_ch1) / (1 + self.atn_peak[i] ** 2))
            new_sig = self.audio_signal.make_copy_with_stft_data(stft_sources, verbose=False)
            new_sig = new_sig.apply_mask(self.result_masks[i])
            new_sig.stft_params = self.stft_params
            source_estimate = new_sig.istft(overwrite=True, truncate_to_length=self.audio_signal.signal_length)
            cur_signal = AudioSignal(audio_data_array=source_estimate, sample_rate=self.sample_rate)
            signals.append(cur_signal)
        return signals

    def plot(self, output_name, three_d_plot=False, normalize=True):
        """Plots histograms with the results of the DUET algorithm

        Parameters:
            output_name (str): path to save plot as
            three_d_plot (Optional[bool]): Flags whether or not to plot in 3d. Defaults to False
            normalize (Optional[bool]): Flags whether the matrix should be normalized or not
        """
        plt.close('all')

        histogram_data = self.get_atn_delay_histogram(recompute=True, normalized=normalize) \
            if self.attenuation_delay_histogram is None \
            else self.attenuation_delay_histogram

        atn_tile = np.tile(self.attenuation_bins[1:], (self.num_delay_bins, 1)).T
        delay_tile = np.tile(self.delay_bins[1:].T, (self.num_attenuation_bins, 1))

        # plot the histogram in 2D
        if not three_d_plot:
            plt.figure()
            plt.pcolormesh(atn_tile, delay_tile, histogram_data)
            plt.xlabel(r'$\alpha$', fontsize=16)
            plt.ylabel(r'$\delta$', fontsize=16)
            plt.title(r'$\alpha-\delta$ Histogram')
            plt.axis('tight')
            plt.savefig(output_name)
            plt.close()

        else:
            # plot the histogram in 3D
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax.plot_wireframe(atn_tile, delay_tile, histogram_data, rstride=2, cstride=2)
            plt.xlabel(r'$\alpha$', fontsize=16)
            plt.ylabel(r'$\delta$', fontsize=16)
            plt.title(r'$\alpha-\delta$ Histogram')
            plt.axis('tight')
            ax.view_init(30, 30)
            plt.savefig(output_name)
            plt.close()
